"""
Network Manager plugin.
"""

from __future__ import annotations

__kupfer_name__ = _("NetworkManager")
__kupfer_sources__ = ("DevicesSource",)
__kupfer_actions__ = ("ToggleWireless",)
__description__ = _("Manage NetworkManager connections")
__version__ = "2023.01"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import time
import typing as ty

import dbus

from kupfer import icons, plugin_support
from kupfer.obj import Action, Leaf, NotAvailableError, Source
from kupfer.support import pretty, weaklib
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_dbus_connection()

_NM_SERVICE = "org.freedesktop.NetworkManager"
_NM_OBJECT = "/org/freedesktop/NetworkManager"
_NM_IFACE = "org.freedesktop.NetworkManager"

_DEVICE_IFACE = "org.freedesktop.NetworkManager.Device"
_PROPS_IFACE = "org.freedesktop.DBus.Properties"
_CONNECTION_IFACE = "org.freedesktop.NetworkManager.Settings.Connection"

_SETTINGS_OBJECT = "/org/freedesktop/NetworkManager/Settings"
_SETTINGS_IFACE = "org.freedesktop.NetworkManager.Settings"


def _create_dbus_connection(iface, obj, service, /, sbus=None):
    """Create dbus connection to NetworkManager"""
    try:
        sbus = sbus or dbus.SystemBus()
        if dobj := sbus.get_object(service, obj):
            return dbus.Interface(dobj, iface)

    except dbus.exceptions.DBusException as err:
        pretty.print_debug(__name__, err)

    raise NotAvailableError(_("NetworkManager"))


def _create_dbus_connection_nm(sbus=None):
    return _create_dbus_connection(
        _NM_IFACE, _NM_OBJECT, _NM_SERVICE, sbus=sbus
    )


def _create_dbus_connection_device(obj, /, sbus=None):
    return _create_dbus_connection(_DEVICE_IFACE, obj, _NM_SERVICE, sbus=sbus)


_NM_DEVICE_STATE: ty.Final[dict[int, str]] = {
    # TRANS: network device status
    0: _("unknown"),
    # TRANS: network device status
    10: _("unmanaged"),
    # TRANS: network device status
    20: _("unavailable"),
    # TRANS: network device status
    30: _("disconnected"),
    # TRANS: network device status
    40: _("prepare"),
    # TRANS: network device status
    50: _("config"),
    # TRANS: network device status
    60: _("need auth"),
    # TRANS: network device status
    70: _("ip config"),
    # TRANS: network device status
    80: _("ip check"),
    # TRANS: network device status
    90: _("secondaries"),
    # TRANS: network device status
    100: _("activated"),
    # TRANS: network device status
    110: _("deactivating"),
    # TRANS: network device status
    120: _("failed"),
}

_DEVICE_ACTIVATED: ty.Final[int] = 100

_NM_DEVICE_TYPES: ty.Final[dict[int, str]] = {
    0: "unknown",
    1: "ethernet",
    2: "wifi",
    3: "unused1",
    4: "unused2",
    5: "bt",
    6: "olpc mesh",
    7: "wimax",
    8: "modem",
    9: "infiniband",
    10: "bond",
    11: "vlan",
    12: "adsl",
    13: "bridge",
    14: "generic",
    15: "team",
    16: "tun",
    17: "ip tunnel",
    18: "macvlan",
    19: "vxlan",
    20: "veth",
    21: "macsec",
    22: "dummy",
    23: "ppp",
    24: "ovs interface",
    25: "ovs port",
    26: "ovs bridge",
    27: "wpan",
    28: "6lowpan",
    29: "wireguard",
    30: "wifi_p2p",
    31: "vrf",
}

_DEV_TYPE_WIRELESS: ty.Final[int] = 2
_DEV_TYPE_VPN: ty.Final[int] = 29


class Device(Leaf):
    def __init__(
        self, path: str, name: str, status: int, managed: bool, devtype: int
    ):
        Leaf.__init__(self, path, name)
        self._status = status
        self.managed = managed
        self.devtype = devtype

    def get_description(self):
        return _("Network device %(dtype)s; state: %(state)s") % {
            "dtype": _NM_DEVICE_TYPES.get(self.devtype) or "unknown",
            "state": _NM_DEVICE_STATE.get(self._status) or str(self._status),
        }

    def get_icon_name(self):
        if self.devtype == _DEV_TYPE_WIRELESS:
            return "network-wireless"

        if self.devtype == _DEV_TYPE_VPN:
            return "network-vpn"

        return "network-wired"

    def get_actions(self):
        yield Disconnect()
        yield Connect()
        yield ShowInfo()

    def status(self) -> int:
        conn = _create_dbus_connection(_PROPS_IFACE, self.object, _NM_SERVICE)
        self._status = int(conn.Get(_DEVICE_IFACE, "State"))
        return self._status


class Disconnect(Action):
    def __init__(self):
        Action.__init__(self, _("Disconnect"))

    def activate(self, leaf, iobj=None, ctx=None):
        bus = dbus.SystemBus()
        try:
            interface = _create_dbus_connection_device(leaf.object, sbus=bus)
            interface.Disconnect()
            time.sleep(1)
        except Exception:
            pretty.print_exc(__name__)

        # return leaf with updated status
        leaf.status()
        return leaf

    def get_icon_name(self):
        return "disconnect"

    def get_description(self):
        return _("Disconnect connection")

    def valid_for_item(self, leaf):
        return leaf.status() == _DEVICE_ACTIVATED

    def has_result(self):
        return True


class Connect(Action):
    def __init__(self):
        # TRANS: activate connection (connect)
        Action.__init__(self, _("Connect..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj

        bus = dbus.SystemBus()
        try:
            interface = _create_dbus_connection_nm(sbus=bus)
            interface.ActivateConnection(iobj.object, leaf.object, "/")
            time.sleep(1)
        except Exception:
            pretty.print_exc(__name__)

        leaf.status()
        return leaf

    def get_description(self):
        return _("Activate connection")

    def get_icon_name(self):
        return "connect"

    def requires_object(self):
        return True

    def object_types(self):
        yield Connection

    def object_source(self, for_item=None):
        return ConnectionsSource(for_item.object, for_item.name)

    def valid_for_item(self, leaf):
        return leaf.status() != _DEVICE_ACTIVATED

    def has_result(self):
        return True


def _get_info_recursive(item, level=0):
    prefix = "    " * level
    if isinstance(item, (dict, dbus.Dictionary)):
        for key, val in item.items():
            yield (f"{prefix}{key}:")
            yield from _get_info_recursive(val, level + 1)

    elif isinstance(item, (tuple, list, dbus.Array)):
        for val in item:
            yield from _get_info_recursive(val)

    elif level > 0:
        # skip garbage on first level
        yield (f"{prefix}{item}")


class ShowInfo(Action):
    def __init__(self):
        Action.__init__(self, _("Show information"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        conn_info = ""
        props_info = ""
        bus = dbus.SystemBus()
        try:
            interface = _create_dbus_connection_device(leaf.object, sbus=bus)
            info = interface.GetAppliedConnection(0)
        except Exception as err:
            conn_info = f"Error: {err}"
        else:
            conn_info = "\n".join(_get_info_recursive(info))

        try:
            interface = _create_dbus_connection(
                _PROPS_IFACE, leaf.object, _NM_SERVICE, sbus=bus
            )
            props = interface.GetAll(_DEVICE_IFACE)
        except Exception as err:
            props_info = f"Error: {err}"
        else:
            props_info = "\n".join(_get_info_recursive(props))

        msg = f"DEVICE\n{props_info}\n------------\n\nCONNECTION\n{conn_info}"
        uiutils.show_text_result(msg, title=_("Connection details"), ctx=ctx)

    def get_description(self):
        return _("Show information about device")


class Connection(Leaf):
    def __init__(self, path: str, name: str, descr: str):
        Leaf.__init__(self, path, name)
        self.descr = descr

    def get_description(self):
        return self.descr

    @staticmethod
    def from_setting(conn: str, settings: dict[str, ty.Any]) -> Connection:
        conn_id = str(settings["id"])
        conn_type = str(settings["type"])
        return Connection(conn, conn_id, conn_type)


class ConnectionsSource(Source):
    source_use_cache = False

    def __init__(self, device_path, interface):
        super().__init__(_("Connections"))
        self.device = device_path
        self.interface = interface

    def get_items(self):
        sbus = dbus.SystemBus()
        dconn = _create_dbus_connection(
            _PROPS_IFACE, self.device, _NM_SERVICE, sbus=sbus
        )
        if not dconn:
            return

        # get available connection for device
        need_check_conn = False
        connections = dconn.Get(_DEVICE_IFACE, "AvailableConnections")
        if not connections:
            # no connections for given device, check all
            dconn = _create_dbus_connection(
                _SETTINGS_IFACE, _SETTINGS_OBJECT, _NM_SERVICE, sbus=sbus
            )
            if not dconn:
                return

            need_check_conn = True
            connections = dconn.ListConnections()

        for conn in connections:
            cset = _create_dbus_connection(
                _CONNECTION_IFACE, conn, _NM_SERVICE, sbus=sbus
            )
            settings = cset.GetSettings()
            settings_connection = settings.get("connection")
            if need_check_conn:
                iface_name = str(settings_connection.get("interface-name"))
                if iface_name != self.interface:
                    continue

            yield Connection.from_setting(conn, settings_connection)


class DevicesSource(Source):
    source_scan_interval: int = 3600

    def __init__(self, name=None):
        Source.__init__(self, name or __kupfer_name__)

    def initialize(self):
        bus = dbus.SystemBus()
        weaklib.dbus_signal_connect_weakly(
            bus,
            "StateChanged",
            self._on_nm_updated,
            dbus_interface=_NM_IFACE,
        )
        weaklib.dbus_signal_connect_weakly(
            bus,
            "DeviceAdded",
            self._on_nm_updated,
            dbus_interface=_NM_IFACE,
        )
        weaklib.dbus_signal_connect_weakly(
            bus,
            "DeviceRemoved",
            self._on_nm_updated,
            dbus_interface=_NM_IFACE,
        )

    def _on_nm_updated(self, *args):
        self.mark_for_update()

    def get_items(self):
        sbus = dbus.SystemBus()
        if interface := _create_dbus_connection_nm(sbus=sbus):
            for dev in interface.GetAllDevices():
                if conn := _create_dbus_connection(
                    _PROPS_IFACE, dev, _NM_SERVICE, sbus=sbus
                ):
                    yield Device(
                        str(dev),
                        str(conn.Get(_DEVICE_IFACE, "Interface")),
                        int(conn.Get(_DEVICE_IFACE, "State")),
                        bool(conn.Get(_DEVICE_IFACE, "Managed")),
                        int(conn.Get(_DEVICE_IFACE, "DeviceType")),
                    )

    def provides(self):
        yield Device

    def get_icon_name(self):
        return "network-wired"


class ToggleWireless(Action):
    def __init__(self):
        Action.__init__(self, "Toggle wireless")

    def _activate(self, sbus: dbus.Bus) -> str | None:
        if interface := _create_dbus_connection(
            _PROPS_IFACE, _NM_OBJECT, _NM_SERVICE, sbus=sbus
        ):
            if not bool(interface.Get(_NM_IFACE, "WirelessHardwareEnabled")):
                # TRANS: notification text when wireless is disabled by hardware
                return _("Hardware wireless disabled")

            state = not bool(interface.Get(_NM_IFACE, "WirelessEnabled"))
            interface.Set(_NM_IFACE, "WirelessEnabled", state)
            if state:
                # TRANS: notification text after wireless enabled
                return _("Wireless enabled")

            # TRANS: notification text after wireless disabled
            return _("Wireless disabled")

        return None

    def activate(self, leaf, iobj=None, ctx=None):
        sbus = dbus.SystemBus()
        if msg := self._activate(sbus):
            uiutils.show_notification("Kupfer", msg)

    def get_description(self):
        return "Toggle wireless by NetworkManager"

    def get_gicon(self):
        return icons.ComposedIcon("network-wireless", "emblem-system")

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf: Leaf) -> bool:
        return bool(leaf.object) and isinstance(leaf.object, DevicesSource)



__kupfer_name__ = _("GNU Screen")
__kupfer_sources__ = ("ScreenSessionsSource",)
__description__ = _("Active GNU Screen sessions")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import pwd
import typing as ty
from pathlib import Path

from kupfer import launch, plugin_support
from kupfer.obj import Action, Leaf, Source
from kupfer.obj.helplib import FilesystemWatchMixin

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_command_available("screen")


def screen_sessions_infos():
    """
    Yield tuples of pid, name, time, status
    for running screen sessions
    """
    with os.popen("screen -list") as pipe:
        output = pipe.read()

    for line in output.splitlines():
        fields = line.split("\t")
        if len(fields) == 4:  # noqa:PLR2004
            _empty, pidname, time, status = fields
            pid, name = pidname.split(".", 1)
            time = time.strip("()")
            status = status.strip("()")
            yield (pid, name, time, status)


def get_username():
    """Return username for current user"""
    info = pwd.getpwuid(os.geteuid())
    return info[0]


class ScreenSession(Leaf):
    """Represented object is the session pid as string"""

    def get_actions(self):
        return (AttachScreen(),)

    def is_valid(self):
        for pid, *_rest in screen_sessions_infos():
            if self.object == pid:
                return True

        return False

    def get_description(self):
        for pid, _name, time, status in screen_sessions_infos():
            if self.object == pid:
                # Handle localization of status
                status_human = status
                if status == "Attached":
                    status_human = _("Attached")
                elif status == "Detached":
                    status_human = _("Detached")

                return _("%(status)s session (%(pid)s) created %(time)s") % {
                    "status": status_human,
                    "pid": pid,
                    "time": time,
                }

        return f"{self.name} ({self.object})"

    def get_icon_name(self):
        return "gnome-window-manager"


class ScreenSessionsSource(Source, FilesystemWatchMixin):
    """Source for GNU Screen sessions"""

    source_scan_interval: int = 3600

    def __init__(self):
        super().__init__(_("Screen Sessions"))
        self.screen_dir = None
        self.monitor_token = None

    def initialize(self):
        ## the screen dir might not exist when we start
        ## luckily, gio can monitor directories before they exist
        self.screen_dir = (
            os.getenv("SCREENDIR") or f"/var/run/screen/S-{get_username()}"
        )
        if not Path(self.screen_dir).exists():
            self.output_debug("Screen socket dir or SCREENDIR not found")

        self.monitor_token = self.monitor_directories(
            self.screen_dir, force=True
        )

    def get_items(self):
        assert self.screen_dir
        if not Path(self.screen_dir).exists():
            return

        for pid, name, _time, _status in screen_sessions_infos():
            yield ScreenSession(pid, name)

    def get_description(self):
        return _("Active GNU Screen sessions")

    def get_icon_name(self):
        return "terminal"

    def provides(self):
        yield ScreenSession


class AttachScreen(Action):
    def __init__(self):
        super().__init__(name=_("Attach"))

    def activate(self, leaf, iobj=None, ctx=None):
        pid = leaf.object
        action_argv = ["screen", "-x", "-R", str(pid)]
        launch.spawn_in_terminal(action_argv)



__kupfer_name__ = _("Claws Mail")
__kupfer_sources__ = ("ClawsContactsSource",)
__kupfer_actions__ = ("NewMailAction", "SendFileByMail")
__description__ = _("Claws Mail Contacts and Actions")
__version__ = "2018-10-07"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import typing as ty
import xml
import xml.parsers
from pathlib import Path
from xml.dom import minidom

from kupfer import launch, plugin_support
from kupfer.obj import Action, FileLeaf, RunnableLeaf, TextLeaf, UrlLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.contacts import ContactLeaf, EmailContact, email_from_leaf
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.obj.helplib import FilesystemWatchMixin

if ty.TYPE_CHECKING:
    from gettext import gettext as _


plugin_support.check_command_available("claws-mail")


class ComposeMail(RunnableLeaf):
    """Create new mail without recipient"""

    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Compose New Email"))

    def run(self, ctx=None):
        launch.spawn_async(["claws-mail", "--compose"])

    def get_description(self):
        return _("Compose a new message in Claws Mail")

    def get_icon_name(self):
        return "mail-message-new"


class ReceiveMail(RunnableLeaf):
    """Receive all new mail from all accounts"""

    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Receive All Email"))

    def run(self, ctx=None):
        launch.spawn_async(["claws-mail", "--receive-all"])

    def get_description(self):
        return _("Receive new messages from all accounts in ClawsMail")

    def get_icon_name(self):
        return "mail-send-receive"


class NewMailAction(Action):
    """Create new mail to selected leaf"""

    def __init__(self):
        Action.__init__(self, _("Compose Email"))

    def activate(self, leaf, iobj=None, ctx=None):
        self.activate_multiple((leaf,))

    def activate_multiple(self, objects):
        recipients = ",".join(filter(None, map(email_from_leaf, objects)))
        launch.spawn_async(["claws-mail", "--compose", recipients])

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_for_item(self, leaf):
        return bool(email_from_leaf(leaf))


class SendFileByMail(Action):
    """Create new e-mail and attach selected file"""

    def __init__(self):
        Action.__init__(self, _("Send in Email To..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        self.activate_multiple((leaf,), (iobj,))

    def activate_multiple(self, objects, iobjects):
        recipients = ",".join(filter(None, map(email_from_leaf, iobjects)))
        attachlist = ["--attach"] + [L.object for L in objects]
        launch.spawn_async(
            ["claws-mail", "--compose", recipients, *attachlist]
        )

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return not leaf.is_dir()

    def requires_object(self):
        return True

    def object_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_object(self, iobj, for_item=None):
        return bool(email_from_leaf(iobj))

    def get_description(self):
        return _("Compose new message in Claws Mail and attach file")

    def get_icon_name(self):
        return "document-send"


class ClawsContactsSource(
    AppLeafContentMixin, ToplevelGroupingSource, FilesystemWatchMixin
):
    appleaf_content_id = "claws-mail"
    source_scan_interval: int = 3600

    def __init__(self, name=_("Claws Mail Address Book")):
        super().__init__(name, "Contacts")
        self._claws_addrbook_dir = Path("~/.claws-mail/addrbook").expanduser()
        self._claws_addrbook_index = self._claws_addrbook_dir.joinpath(
            "addrbook--index.xml"
        )
        self._version = 5
        self.monitor_token = None

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        if not self._claws_addrbook_dir.is_dir():
            return

        self.monitor_token = self.monitor_directories(
            str(self._claws_addrbook_dir)
        )

    def monitor_include_file(self, gfile):
        # monitor only addrbook-*.xml files
        return (
            gfile
            and gfile.get_basename().endswith(".xml")
            and gfile.get_basename().startswith("addrbook-")
        )

    def mark_for_update(self, postpone=False):
        super().mark_for_update(postpone=True)

    def get_items(self):
        if self._claws_addrbook_index.is_file():
            for addrbook_file in self._load_address_books():
                addrbook_filepath = self._claws_addrbook_dir.joinpath(
                    addrbook_file
                )
                if not addrbook_filepath.exists():
                    continue

                try:
                    dtree = minidom.parse(str(addrbook_filepath))
                    persons = dtree.getElementsByTagName("person")
                    for person in persons:
                        commonname = person.getAttribute("cn")
                        addresses = person.getElementsByTagName("address")
                        for address in addresses:
                            email = address.getAttribute("email")
                            yield EmailContact(email, commonname)

                except (Exception, xml.parsers.expat.ExpatError) as err:
                    self.output_error(err)

        yield ComposeMail()
        yield ReceiveMail()

    def should_sort_lexically(self):
        # since it is a grouping source, grouping and non-grouping will be
        # separate and only grouping leaves will be sorted
        return True

    def get_description(self):
        return _("Contacts from Claws Mail Address Book")

    def get_icon_name(self):
        return "claws-mail"

    def provides(self):
        yield RunnableLeaf
        yield ContactLeaf

    def _load_address_books(self):
        """load list of address-book files"""
        try:
            dtree = minidom.parse(str(self._claws_addrbook_index))
            for book in dtree.getElementsByTagName("book"):
                yield book.getAttribute("file")

        except (Exception, xml.parsers.expat.ExpatError) as err:
            self.output_error(err)



"""
Accelerators configuration.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import json
import typing as ty

from kupfer import config
from kupfer.support import pretty

_repr_key = repr


__all__ = ("AccelConfig",)

# AcceleratorValidator check is given accelerator is valid
AcceleratorValidator = ty.Callable[[str], bool]


class AccelConfig(pretty.OutputMixin):
    """Action Accelerator configuration"""

    def __init__(self):
        self._accels: dict[str, str] = {}
        # is configuration loaded
        self._loaded: bool = False
        # is configuration changed and not saved
        self._changed: bool = False

    def _filename(self) -> str | None:
        if ret := config.save_config_file("action_accels.json"):
            return ret

        self.output_error("Can't find XDG_CONFIG_HOME")
        return None

    def load(self, validate_func: AcceleratorValidator) -> bool:
        """Load accelerators configuration.

        If configuration is already loaded - do nothing.

        After loading `validate_func` is used for each accelerator; if function
        return false - accelerator is ignored.

        Return True on success.
        """
        if self._loaded:
            return True

        self._loaded = True
        data_file = self._filename()
        if data_file is None:
            return False

        self._accels.clear()

        try:
            with open(data_file, encoding="UTF-8") as dfp:
                accels = json.load(dfp)

            self.output_debug("Read", data_file)
        except FileNotFoundError:
            return False
        except Exception:
            self.output_error("Failed to read:", data_file)
            self.output_exc()
            return False

        try:
            self._valid_accel(accels, validate_func)
        except Exception:
            self.output_exc()

        self.output_debug("Loaded", self._accels)
        return True

    def _valid_accel(
        self, accels: dict[ty.Any, ty.Any], validate_func: AcceleratorValidator
    ) -> None:
        """Validate loaded data."""
        if not isinstance(self._accels, dict):
            raise TypeError("Accelerators must be a dictionary")

        for obj, key in accels.items():
            # make sure object and key are strings
            key = str(key)  # noqa:PLW2901
            if validate_func(key):
                self._accels[str(obj)] = key
            else:
                self.output_error("Ignoring invalid accel", key, "for", obj)

    def get(self, obj: ty.Any) -> str | None:
        """Return accel key for @obj or None"""
        return self._accels.get(_repr_key(obj))

    def set(self, obj: ty.Any, key: str) -> None:
        """Set accelerator `key` for `obj`."""
        self.output_debug("Set", key, "for", _repr_key(obj))
        assert hasattr(obj, "activate")
        assert isinstance(key, str)
        self._accels[_repr_key(obj)] = key
        self._changed = True

    def store(self) -> None:
        """Write all accelerator into configuration file."""
        if not self._changed:
            return

        data_file = self._filename()
        if data_file is None:
            return

        self.output_debug("Writing to", data_file)
        try:
            with open(data_file, "w", encoding="UTF-8") as dfp:
                json.dump(self._accels, dfp, indent=4, sort_keys=True)

        except Exception:
            self.output_error("Failed to write:", data_file)
            self.output_exc()

        self._changed = False



from __future__ import annotations

import os

try:
    from gi.repository import Wnck
except ImportError:
    Wnck = None

from kupfer import config
from kupfer.support import datatools, pretty

__all__ = ("allows_keybinder", "is_kwin", "is_wayland")


@datatools.evaluate_once
def is_kwin() -> bool:
    """Try to figure out if KWin is the current window manager.
    If Wnck unavailable try guess from environment variables.
    """
    if Wnck and (screen := Wnck.Screen.get_default()) is not None:
        winmgr: str | None = screen.get_window_manager_name()
        pretty.print_debug(__name__, "window manager is", winmgr)
        if winmgr:
            return winmgr.lower() == "kwin"

    return _desktop_environment_guess().lower() == "kde"


def _desktop_environment_guess() -> str:
    ret = os.getenv("XDG_CURRENT_DESKTOP") or ""
    pretty.print_debug(__name__, "desktop environment is", ret)
    return ret


def is_wayland() -> bool:
    """Check is Wayland is current graphics environment"""
    return bool(os.getenv("WAYLAND_DISPLAY"))


def allows_keybinder() -> bool:
    """Check is keybinder available."""
    return config.has_capability("KEYBINDER") and not is_wayland()



from __future__ import annotations

import typing as ty
from contextlib import suppress

from gi.repository import GdkPixbuf, Gio, Gtk
from gi.repository.Gio import (
    FILE_ATTRIBUTE_STANDARD_ICON,
    FILE_ATTRIBUTE_THUMBNAIL_PATH,
    File,
    FileIcon,
    Icon,
    ThemedIcon,
)
from gi.repository.GLib import GError

from kupfer.core import settings
from kupfer.support import datatools, pretty, scheduler

__all__ = (
    "ComposedIcon",
    "ComposedIconSmall",
    "IconRenderer",
    "get_gicon_for_file",
    "get_gicon_for_names",
    "get_gicon_from_file",
    "get_gicon_with_fallbacks",
    "get_good_name_for_icon_names",
    "get_icon_for_gicon",
    "get_icon_for_name",
    "get_icon_from_file",
    "get_pixbuf_from_data",
    "get_pixbuf_from_file",
    "get_thumbnail_for_gfile",
    "is_good_gicon",
    "parse_load_icon_list",
)

_ICON_CACHE: ty.Final[
    datatools.LruCache[tuple[int, str], GdkPixbuf.Pixbuf]
] = datatools.LruCache(128, name="_ICON_CACHE")

_LARGE_SZ = 128
_SMALL_SZ = 24

## default fallbacks for our themable icons
_KUPFER_ICON_FALLBACKS: ty.Final = {
    "kupfer-execute": "system-run",
    "kupfer-object": "text-x-generic",
    "kupfer-object-multiple": "folder",
    "kupfer-catalog": "system-search",
    "kupfer-search": "system-search",
}

_KUPFER_LOCALLY_INSTALLED_NAMES: ty.Final[set[str]] = set()


def _on_icon_theme_changed(theme):
    pretty.print_info(__name__, "Icon theme changed, clearing cache")
    _ICON_CACHE.clear()


_default_theme = Gtk.IconTheme.get_default()  # pylint: disable=no-member
_default_theme.connect("changed", _on_icon_theme_changed)
_local_theme = Gtk.IconTheme.new()
_local_theme.set_search_path([])


def parse_load_icon_list(
    icon_list_data: bytes,
    get_data_func: ty.Callable[[str], bytes],
    plugin_name: str | None = None,
) -> None:
    """
    @icon_list_data: A bytestring whose lines identify icons
    @get_data_func: A function to return the data for a relative filename
    @plugin_name: plugin id, if applicable
    """
    try:
        icon_list_string = icon_list_data.decode("utf-8")
    except UnicodeDecodeError as exc:
        pretty.print_error(
            __name__,
            f"Malformed icon-list in plugin {plugin_name!r}: {exc!r}",
        )
        raise

    for line in icon_list_string.splitlines():
        # ignore '#'-comments
        if line.startswith("#") or not line.strip():
            continue

        fields = tuple(map(str.strip, line.split("\t", 2)))
        if len(fields) <= 1:
            pretty.print_error(
                __name__,
                "Malformed icon-list line {line!r} from {plugin_name!r}",
            )
            continue

        icon_name, basename, *_tail = fields
        override = "!override" in fields
        _load_icon_from_func(
            plugin_name, icon_name, get_data_func, override, func_args=basename
        )


def _load_icon_from_func(
    plugin_name: str | None,
    icon_name: str,
    get_data_func: ty.Callable[[str], bytes],
    override: bool,
    func_args: str,
) -> None:
    """
    Load icon from @icon_data into the name @icon_name

    @get_data_func: function to retrieve the data if needed
    @override: override the icon theme
    """
    if not override and icon_name in _KUPFER_LOCALLY_INSTALLED_NAMES:
        pretty.print_debug(__name__, "Skipping existing", icon_name)
        return

    if not override and _default_theme.has_icon(icon_name):
        pretty.print_debug(__name__, "Skipping themed icon", icon_name)
        return

    try:
        icon_data = get_data_func(func_args)
    except Exception:
        pretty.print_error(
            __name__, f"Error loading icon {icon_name!r} for {plugin_name!r}"
        )
        pretty.print_exc(__name__)
        return

    for size in (_SMALL_SZ, _LARGE_SZ):
        pixbuf = get_pixbuf_from_data(icon_data, size, size)
        Gtk.IconTheme.add_builtin_icon(  # pylint: disable=no-member
            icon_name, size, pixbuf
        )
        pretty.print_debug(
            __name__, "Loading icon", icon_name, "at", size, "for", plugin_name
        )

    _KUPFER_LOCALLY_INSTALLED_NAMES.add(icon_name)


def _get_icon(key: str, icon_size: int) -> ty.Iterator[GdkPixbuf.Pixbuf]:
    """
    try retrieve icon in cache
    is a generator so it can be concisely called with a for loop

    FIXME: not used (?)
    """
    with suppress(KeyError):
        yield _ICON_CACHE[(icon_size, key)]


def _store_icon(key: str, icon_size: int, icon: GdkPixbuf.Pixbuf) -> None:
    """
    Store an icon in cache. It must not have been stored before
    """
    assert icon, f"icon {key} may not be {icon}"
    _ICON_CACHE[(icon_size, key)] = icon


def _get_icon_dwim(
    icon: ty.Union[Icon, str], icon_size: int
) -> GdkPixbuf.Pixbuf | None:
    """Make an icon at @icon_size where
    @icon can be either an icon name, or a gicon
    """
    if isinstance(icon, Icon):
        return get_icon_for_gicon(icon, icon_size)

    if icon:
        return get_icon_for_name(icon, icon_size)

    pretty.print_debug(__name__, "_get_icon_dwim failed for", icon, icon_size)
    return None


# pylint: disable=too-few-public-methods
class ComposedIcon:
    """
    A composed icon, which kupfer will render to pixbuf as
    background icon with the decorating icon as emblem

    If `emblem_is_fallback` is True, use emblem if rendered icon is smaller
    than `minimum_icon_size`.

    ComposedIcon is cached, i.e. when ComposedIcon is created first cache is
    checked if there is already object with the same parameters. If yes - its
    returned, otherwise new object is created.

    Icon itself is rendered only once.
    """

    _cache: datatools.LruCache[tuple[ty.Any, ...], GdkPixbuf.Pixbuf | None] = (
        datatools.LruCache(32, name="ComposedIcon cache")
    )

    __slots__ = (
        "_rendered",
        "_rendered_size",
        "baseicon",
        "emblem_is_fallback",
        "emblemicon",
        "minimum_icon_size",
    )

    def __init__(
        self,
        baseicon: GIcon | str,
        emblem: GIcon | str,
        emblem_is_fallback: bool = False,
        minimum_icon_size: int = 48,
    ) -> None:
        self.minimum_icon_size = minimum_icon_size
        self.baseicon = baseicon
        self.emblemicon = emblem
        self.emblem_is_fallback = emblem_is_fallback

    @classmethod
    def new(cls, *args, **kwargs):
        """Construct a composed icon from @baseicon and @emblem,
        which may be GIcons or icon names (strings)
        """
        return cls(*args, **kwargs)

    def render_composed_icon(self, icon_size: int) -> GdkPixbuf.Pixbuf | None:
        key = (
            self.baseicon,
            self.emblemicon,
            self.emblem_is_fallback,
            self.minimum_icon_size,
            icon_size,
        )
        with suppress(KeyError):
            return ComposedIcon._cache[key]

        # If it's too small, render as fallback icon
        if icon_size < self.minimum_icon_size:
            if self.emblem_is_fallback:
                icon = _get_icon_for_standard_gicon(self.emblemicon, icon_size)
            else:
                icon = _get_icon_for_standard_gicon(self.baseicon, icon_size)

            ComposedIcon._cache[key] = icon
            return icon

        toppbuf = _get_icon_dwim(self.emblemicon, icon_size)
        bottompbuf = _get_icon_dwim(self.baseicon, icon_size)
        if not toppbuf or not bottompbuf:
            ComposedIcon._cache[key] = None
            return None

        dest = bottompbuf.copy()
        # @frac is the scale
        frac = 0.6
        dcoord = int((1 - frac) * icon_size)
        dsize = int(frac * icon_size)
        # http://library.gnome.org/devel/gdk-pixbuf/unstable//gdk-pixbuf-scaling.html
        toppbuf.composite(
            dest,
            dcoord,
            dcoord,
            dsize,
            dsize,
            dcoord,
            dcoord,
            frac,
            frac,
            GdkPixbuf.InterpType.BILINEAR,
            255,
        )

        ComposedIcon._cache[key] = dest
        return dest


# pylint: disable=invalid-name
def ComposedIconSmall(  # noqa:N802
    baseicon: GIcon | str, emblem: GIcon | str, **kwargs: ty.Any
) -> ComposedIcon:
    """Create composed icon for leaves with emblem visible on browser list"""
    return ComposedIcon(
        baseicon, emblem, minimum_icon_size=_SMALL_SZ, **kwargs
    )


GIcon = ty.Union[ComposedIcon, ThemedIcon, FileIcon]


def get_thumbnail_for_gfile(
    gfile: Gio.File, width: int = -1, height: int = -1
) -> GdkPixbuf.Pixbuf | None:
    """
    Return a Pixbuf thumbnail for the file at
    gfile: Gio.File
    size is @width x @height

    return None if not found
    """
    if not gfile.query_exists():
        return None

    finfo = gfile.query_info(
        FILE_ATTRIBUTE_THUMBNAIL_PATH, Gio.FileQueryInfoFlags.NONE, None
    )
    thumb_path = finfo.get_attribute_byte_string(FILE_ATTRIBUTE_THUMBNAIL_PATH)

    return get_pixbuf_from_file(thumb_path, width, height)


def get_pixbuf_from_file(
    thumb_path: str | None, width: int = -1, height: int = -1
) -> GdkPixbuf.Pixbuf | None:
    """
    Return a Pixbuf thumbnail for the file at @thumb_path
    sized @width x @height
    For non-icon pixbufs:
    We might cache these, but on different terms than the icon cache
    if @thumb_path is None, return None
    """
    if not thumb_path:
        return None
    try:
        return GdkPixbuf.Pixbuf.new_from_file_at_size(
            thumb_path, width, height
        )
    except GError as exc:
        # this error is not important, the program continues on fine,
        # so we put it in debug output.
        pretty.print_debug(
            __name__, "get_pixbuf_from_file file:", thumb_path, "error:", exc
        )

    return None


_GICON_CACHE: datatools.LruCache[str, GIcon | None] = datatools.LruCache(
    128, name="_GICON_CACHE"
)


def get_gicon_for_file(uri: str) -> GIcon | None:
    """
    Return a GIcon representing the file at
    the @uri, which can be *either* and uri or a path

    return None if not found
    """

    with suppress(KeyError):
        return _GICON_CACHE[uri]

    gfile = File.new_for_path(uri)
    if not gfile.query_exists():
        gfile = File.new_for_uri(uri)
        if not gfile.query_exists():
            pretty.print_debug(__name__, "get_gicon_for_file", uri, "failed")
            _GICON_CACHE[uri] = None
            return None

    finfo = gfile.query_info(
        FILE_ATTRIBUTE_STANDARD_ICON, Gio.FileQueryInfoFlags.NONE, None
    )
    gicon = finfo.get_attribute_object(FILE_ATTRIBUTE_STANDARD_ICON)
    _GICON_CACHE[uri] = gicon
    return gicon


def get_gicon_from_file(path: str) -> FileIcon | None:
    """Load GIcon from @path; return None if failed."""
    with suppress(KeyError):
        return _GICON_CACHE[path]

    icon = None
    file = File.new_for_path(path)
    if file.query_exists():
        icon = FileIcon.new(file)
    else:
        pretty.print_debug(__name__, "get_gicon_from_file", path, "failed")

    _GICON_CACHE[path] = icon
    return icon


def get_icon_for_gicon(
    gicon: GIcon, icon_size: int
) -> GdkPixbuf.Pixbuf | None:
    """
    Return a pixbuf of @icon_size for the @gicon

    NOTE: Currently only the following can be rendered:
        Gio.ThemedIcon
        Gio.FileIcon
        kupfer.icons.ComposedIcon
    """
    # FIXME: We can't load any general GIcon
    if not gicon:
        return None

    if isinstance(gicon, ComposedIcon):
        return gicon.render_composed_icon(icon_size)

    return _get_icon_for_standard_gicon(gicon, icon_size)


def _get_icon_for_standard_gicon(
    gicon: ty.Any, icon_size: int
) -> GdkPixbuf.Pixbuf | None:
    """Render ThemedIcon and FileIcon"""
    if isinstance(gicon, FileIcon):
        ifile = gicon.get_file()
        return get_icon_from_file(ifile.get_path(), icon_size)

    if isinstance(gicon, ThemedIcon):
        names = gicon.get_names()
        return get_icon_for_name(names[0], icon_size, names)

    pretty.print_debug(
        __name__, "_get_icon_for_standard_gicon, could not load", gicon
    )
    return None


class IconRenderer:
    """
    Default GTK+ implementation
    """

    @classmethod
    def pixbuf_for_name(
        cls, icon_name: str, icon_size: int
    ) -> GdkPixbuf.Pixbuf | None:
        if icon_name in _KUPFER_LOCALLY_INSTALLED_NAMES:
            with suppress(GError):
                return _local_theme.load_icon(
                    icon_name,
                    icon_size,
                    Gtk.IconLookupFlags.USE_BUILTIN  # pylint: disable=no-member
                    | Gtk.IconLookupFlags.FORCE_SIZE,  # pylint: disable=no-member
                )

        with suppress(GError):
            return _default_theme.load_icon(
                icon_name,
                icon_size,
                Gtk.IconLookupFlags.USE_BUILTIN  # pylint: disable=no-member
                | Gtk.IconLookupFlags.FORCE_SIZE,  # pylint: disable=no-member
            )

        return None

    @classmethod
    def pixbuf_for_file(
        cls, file_path: str, icon_size: int
    ) -> GdkPixbuf.Pixbuf | None:
        with suppress(GError):
            return GdkPixbuf.Pixbuf.new_from_file_at_size(
                file_path, icon_size, icon_size
            )

        return None


_ICON_RENDERER = IconRenderer  # pylint: disable=invalid-name


def _setup_icon_renderer(_sched: ty.Any) -> None:
    setctl = settings.get_settings_controller()
    setctl.connect(
        "alternatives-changed::icon_renderer", _on_icon_render_change
    )
    setctl.connect(
        "value-changed::tools.icon_renderer", _on_icon_render_change
    )
    _on_icon_render_change(setctl)


def _on_icon_render_change(setctl, *_arguments):
    global _ICON_RENDERER  # pylint: disable=global-statement # noqa: PLW0603
    renderer_dict = setctl.get_preferred_alternative("icon_renderer")
    assert renderer_dict

    renderer = renderer_dict.get("renderer")
    if not renderer or renderer is _ICON_RENDERER:
        return

    pretty.print_debug(__name__, "Using", renderer)
    _on_icon_theme_changed(None)
    _ICON_RENDERER = renderer


scheduler.get_scheduler().connect("loaded", _setup_icon_renderer)


def get_icon_for_name(
    icon_name: str,
    icon_size: int,
    icon_names: ty.Iterable[str] | None = None,
) -> GdkPixbuf.Pixbuf | None:
    if icon_name:
        with suppress(KeyError):
            return _ICON_CACHE[(icon_size, icon_name)]

    # Try the whole list of given names
    for load_name in icon_names or (icon_name,):
        try:
            if icon := _ICON_RENDERER.pixbuf_for_name(load_name, icon_size):
                break

            if (fallback_name := _KUPFER_ICON_FALLBACKS.get(icon_name)) and (
                icon := _ICON_RENDERER.pixbuf_for_name(
                    fallback_name, icon_size
                )
            ):
                break

        except Exception:
            pretty.print_exc(__name__)
            icon = None
    else:
        # if we did not reach 'break' in the loop
        pretty.print_debug(
            __name__,
            "get_gicon_for_name name",
            icon_name,
            "size",
            icon_size,
            "names",
            icon_names,
            "failed",
        )
        return None

    # We store the first icon in the list, even if the match
    # found was later in the chain
    _store_icon(icon_name, icon_size, icon)
    return icon


def get_icon_from_file(
    icon_file: str, icon_size: int
) -> GdkPixbuf.Pixbuf | None:
    # try to load from cache
    with suppress(KeyError):
        return _ICON_CACHE[(icon_size, icon_file)]

    if icon := _ICON_RENDERER.pixbuf_for_file(icon_file, icon_size):
        _store_icon(icon_file, icon_size, icon)
        return icon

    return None


def is_good_gicon(gicon: GIcon | None) -> bool:
    """Return True if it is likely that @gicon will load a visible icon
    (icon name exists in theme, or icon references existing file)
    """
    if not gicon:
        return False

    if isinstance(gicon, ThemedIcon):
        return bool(get_good_name_for_icon_names(gicon.get_names()))

    if isinstance(gicon, FileIcon):
        ifile = gicon.get_file()
        return ifile.query_exists()  # type: ignore

    # Since we can't load it otherwise (right now, see above)
    return False


def get_gicon_with_fallbacks(
    gicon: GIcon | None, names: ty.Iterable[str]
) -> GIcon | None:
    if is_good_gicon(gicon):
        return gicon

    name = None
    for name in names:
        gicon = ThemedIcon.new(name)
        if is_good_gicon(gicon):
            return gicon

    return ThemedIcon.new(name) if name else None


def get_good_name_for_icon_names(names: ty.Iterable[str]) -> str | None:
    """Return first name in @names that exists
    in current icon theme, or None
    """
    for name in names:
        if _default_theme.has_icon(name):
            return name

    return None


def get_gicon_for_names(*names: str) -> ThemedIcon:
    return ThemedIcon.new_from_names(names)


def get_pixbuf_from_data(
    data: bytes,
    width: int | None = None,
    height: int | None = None,
) -> GdkPixbuf.Pixbuf:
    """Create pixbuf object from data with optional scaling

    @data: picture as raw data
    @width, @height: optional destination size
    """

    ploader = GdkPixbuf.PixbufLoader.new()

    if width and height:

        def set_size(
            img: GdkPixbuf.PixbufLoader, img_width: int, img_height: int
        ) -> None:
            assert width and height
            scale = min(width / float(img_width), height / float(img_height))
            new_width, new_height = (
                int(img_width * scale),
                int(img_height * scale),
            )
            img.set_size(new_width, new_height)

        ploader.connect("size-prepared", set_size)

    ploader.write(data)
    ploader.close()
    return ploader.get_pixbuf()



"""
virtualbox_const_support.py

Constants for VirtualBox.
"""

__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"
__version__ = "0.4"

# virtual machine states
VM_STATE_POWEROFF = 0
VM_STATE_POWERON = 1
VM_STATE_PAUSED = 2
VM_STATE_SAVED = 3

# virtual machine actions
VM_START_NORMAL = 1
VM_START_HEADLESS = 2
VM_PAUSE = 3
VM_POWEROFF = 4
VM_ACPI_POWEROFF = 5
VM_REBOOT = 6
VM_RESUME = 7
VM_SAVE = 8



__kupfer_name__ = _("Default Email Client")
__kupfer_actions__ = ("NewMailAction",)
__description__ = _("Compose email using the system's default mailto: handler")
__version__ = "2017.1"
__author__ = ""

import typing as ty

from kupfer import launch
from kupfer.obj import Action, TextLeaf, UrlLeaf
from kupfer.obj.contacts import ContactLeaf, email_from_leaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class NewMailAction(Action):
    def __init__(self):
        Action.__init__(self, _("Compose Email To"))

    def activate(self, leaf, iobj=None, ctx=None):
        email = email_from_leaf(leaf)
        launch.show_url(f"mailto:{email}")

    def activate_multiple(self, objects):
        recipients = ",".join(filter(None, map(email_from_leaf, objects)))
        url = "mailto:" + recipients
        launch.show_url(url)

    def item_types(self):
        yield ContactLeaf
        yield TextLeaf
        yield UrlLeaf

    def valid_for_item(self, leaf):
        return bool(email_from_leaf(leaf))

    def get_description(self):
        return __description__

    def get_icon_name(self):
        return "mail-message-new"



__kupfer_name__ = _("LXQT Session Management")
__kupfer_sources__ = ("LxqtItemsSource",)
__description__ = _("Actions for LXQT desktop")
__version__ = "2020-08-23"
__author__ = "Leonardo Masuero <leom255255@gmail.com>"
# Based on XFCE Session Management by Karol Będkowski
import typing as ty

from kupfer.plugin import session_support as support

if ty.TYPE_CHECKING:
    from gettext import gettext as _

# sequences of argument lists
_LOGOUT_CMD = (["lxqt-leave", "--logout"],)
_SHUTDOWN_CMD = (["lxqt-leave", "--shutdown"],)
_LOCKSCREEN_CMD = (["lxqt-leave", "--lockscreen"],)


class LxqtItemsSource(support.CommonSource):
    source_scan_interval: int = 36000

    def __init__(self):
        support.CommonSource.__init__(self, _("LXQT Session Management"))

    def get_items(self):
        return (
            support.Logout(_LOGOUT_CMD),
            support.LockScreen(_LOCKSCREEN_CMD),
            support.Shutdown(_SHUTDOWN_CMD),
        )



from __future__ import annotations

__kupfer_name__ = _("Document Templates")
__kupfer_sources__ = ("TemplatesSource",)
__kupfer_actions__ = ("CreateNewDocument",)
__description__ = _("Create new documents from your templates")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import typing as ty
from pathlib import Path

from gi.repository import Gio, GLib

from kupfer import icons
from kupfer.obj import Action, FileLeaf, Leaf, Source, helplib
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.support import fileutils, pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

DEFAULT_TMPL_DIR = "~/Templates"


class Template(FileLeaf):
    def __init__(self, path):
        basename = GLib.filename_display_basename(path)
        nameroot, _ext = os.path.splitext(basename)
        FileLeaf.__init__(self, path, _("%s template") % nameroot)

    def get_actions(self):
        yield CreateDocumentIn()
        yield from FileLeaf.get_actions(self)

    def get_gicon(self):
        file_gicon = FileLeaf.get_gicon(self)
        return icons.ComposedIcon("text-x-generic-template", file_gicon)


class EmptyFile(Leaf):
    def __init__(self):
        Leaf.__init__(self, None, _("Empty File"))

    def repr_key(self):
        return None

    def get_actions(self):
        yield CreateDocumentIn()

    def get_icon_name(self):
        return "text-x-generic"


class NewFolder(Leaf):
    def __init__(self):
        Leaf.__init__(self, None, _("New Folder"))

    def repr_key(self):
        return None

    def get_actions(self):
        yield CreateDocumentIn()

    def get_icon_name(self):
        return "folder"


class CreateNewDocument(Action):
    def __init__(self):
        Action.__init__(self, _("Create New Document..."))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        destpath: str | None
        if iobj.object is not None:
            # Copy the template to destination directory
            basename = os.path.basename(iobj.object)
            tmpl_gfile = Gio.File.new_for_path(iobj.object)
            destpath = fileutils.get_destpath_in_directory(
                leaf.object, basename
            )
            destfile = Gio.File.new_for_path(destpath)
            tmpl_gfile.copy(
                destfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None
            )
        elif isinstance(iobj, NewFolder):
            filename = str(iobj)
            destpath = fileutils.get_destpath_in_directory(
                leaf.object, filename
            )
            Path(destpath).mkdir(parents=True)
        else:
            # create new empty file
            filename = str(iobj)
            destfile, destpath = fileutils.get_destfile_in_directory(
                leaf.object, filename
            )
            if destfile:
                destfile.close()

        return FileLeaf(destpath) if destpath else None

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return leaf.is_dir()

    def requires_object(self):
        return True

    def object_types(self):
        yield Template
        yield EmptyFile
        yield NewFolder

    def object_source(self, for_item=None):
        return TemplatesSource()

    def get_description(self):
        return _("Create a new document from template")

    def get_icon_name(self):
        return "document-new"


class CreateDocumentIn(
    helplib.reverse_action(CreateNewDocument)  # type:ignore
):
    rank_adjust = 10

    # pylint: disable=super-init-not-called,non-parent-init-called
    def __init__(self):
        Action.__init__(self, _("Create Document In..."))


def _get_tmpl_dir():
    tmpl_dir = GLib.get_user_special_dir(
        GLib.UserDirectory.DIRECTORY_TEMPLATES
    )
    if tmpl_dir == os.path.expanduser("~"):
        tmpl_dir = None

    if not tmpl_dir:
        tmpl_dir = os.path.expanduser(DEFAULT_TMPL_DIR)

    pretty.print_debug(__name__, tmpl_dir)
    return tmpl_dir


class TemplatesSource(Source, FilesystemWatchMixin):
    source_scan_interval: int = 3600

    def __init__(self):
        Source.__init__(self, _("Document Templates"))
        self.monitor_token = None

    def initialize(self):
        self.monitor_token = self.monitor_directories(_get_tmpl_dir())

    def get_items(self):
        tmpl_dir = _get_tmpl_dir()
        yield EmptyFile()
        yield NewFolder()
        try:
            with os.scandir(tmpl_dir) as entries:
                for entry in entries:
                    yield Template(entry.path)

        except OSError as exc:
            self.output_error(exc)

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return None

    def get_icon_name(self):
        return "system-file-manager"

    def provides(self):
        yield Template



"""
Changes:
    2012-06-21 Karol Będkowski:
        * support new dbus api introduced in GTG 0.2.9
"""

__kupfer_name__ = _("Getting Things GNOME")
__kupfer_sources__ = ("TasksSource",)
__kupfer_actions__ = ("CreateNewTask", "CreateNewEmptyTask")
__description__ = _("Browse and create new tasks in GTG")
__version__ = "2017.2"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>, US"

import typing as ty

import dbus

from kupfer import plugin_support
from kupfer.obj import (
    Action,
    AppLeaf,
    Leaf,
    NotAvailableError,
    Source,
    TextLeaf,
)
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.support import pretty, textutils, weaklib

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_dbus_connection()

_GTG_ID = "gtg"
_SERVICE_NAME2 = "org.gnome.GTG"
_OBJECT_NAME2 = "/org/gnome/GTG"
_IFACE_NAME2 = "org.gnome.GTG"

_STATUS_RANK = {
    s: i for i, s in enumerate(("Active", "Postponed", "Dismiss", "Done"))
}
_STATUS_RANK_DEFAULT = len(_STATUS_RANK)


def _create_dbus_connection_gtg(iface, obj, service, activate=False):
    """Create dbus connection to GTG
    @activate: if True, start program if not running
    """
    interface = None
    sbus = dbus.SessionBus()
    try:
        proxy_obj = sbus.get_object(
            "org.freedesktop.DBus", "/org/freedesktop/DBus"
        )
        dbus_iface = dbus.Interface(proxy_obj, "org.freedesktop.DBus")
        if (activate or dbus_iface.NameHasOwner(iface)) and (
            obj := sbus.get_object(service, obj)
        ):
            interface = dbus.Interface(obj, iface)

    except dbus.exceptions.DBusException as err:
        pretty.print_debug(__name__, err)

    return interface


def _create_dbus_connection(activate=False):
    interface = _create_dbus_connection_gtg(
        _IFACE_NAME2, _OBJECT_NAME2, _SERVICE_NAME2, activate
    )
    if interface is None:
        pretty.print_error(__name__, "Cannot connect to GTG via DBus")
        if activate:
            raise NotAvailableError(_("Getting Things GNOME"))

    return interface


def _truncate_long_text(text, maxlen=80):
    if len(text) > maxlen:
        return text[: maxlen - 1] + "…"

    return text


def _load_tasks(interface):
    """Load task by dbus interface"""
    tasks = interface.GetTasks()
    if not tasks:
        tasks = interface.GetTasksFiltered("")

    for task in tasks:
        title = task["title"].strip()
        if not title:
            title = task["text"].strip()

        title = _truncate_long_text(title)
        otask = Task(task["id"], title, task["status"])
        otask.duedate = task["duedate"]
        otask.startdate = task["startdate"]
        otask.tags = task["tags"]
        yield otask


def _change_task_status(task_id, status):
    interface = _create_dbus_connection(True)
    task = interface.GetTask(task_id)
    task["status"] = status
    interface.ModifyTask(task_id, task)


class Task(Leaf):
    def __init__(self, task_id, title, status):
        Leaf.__init__(self, task_id, title)
        self.status = status
        self.tags = None
        self.duedate = None
        self.startdate = None

    def get_description(self):
        descr = [self.status]
        if self.duedate:
            descr.append(_("due: %s") % self.duedate)

        if self.startdate:
            descr.append(_("start: %s") % self.startdate)

        if self.tags:
            descr.append(_("tags: %s") % " ".join(self.tags))

        return "  ".join(descr)

    def get_icon_name(self):
        return "gtg"

    def get_actions(self):
        yield OpenEditor()
        yield Delete()
        if self.status != "Done":
            yield MarkDone()

        if self.status != "Dismiss":
            yield Dismiss()

    def sort_key(self):
        return _STATUS_RANK.get(self.status, _STATUS_RANK_DEFAULT)


class OpenEditor(Action):
    rank_adjust = 1
    action_accelerator = "o"

    def __init__(self):
        Action.__init__(self, _("Open"))

    def activate(self, leaf, iobj=None, ctx=None):
        interface = _create_dbus_connection(True)
        interface.OpenTaskEditor(leaf.object)

    def get_icon_name(self):
        return "document-open"

    def get_description(self):
        return _("Open task in Getting Things GNOME!")


class Delete(Action):
    rank_adjust = -10

    def __init__(self):
        Action.__init__(self, _("Delete"))

    def activate(self, leaf, iobj=None, ctx=None):
        interface = _create_dbus_connection(True)
        interface.DeleteTask(leaf.object)

    def get_icon_name(self):
        return "edit-delete"

    def get_description(self):
        return _("Permanently remove this task")


class MarkDone(Action):
    action_accelerator = "d"

    def __init__(self):
        Action.__init__(self, _("Mark Done"))

    def activate(self, leaf, iobj=None, ctx=None):
        _change_task_status(leaf.object, "Done")

    def get_icon_name(self):
        return "gtk-yes"

    def get_description(self):
        return _("Mark this task as done")


class Dismiss(Action):
    action_accelerator = "i"

    def __init__(self):
        Action.__init__(self, _("Dismiss"))

    def activate(self, leaf, iobj=None, ctx=None):
        _change_task_status(leaf.object, "Dismiss")

    def get_icon_name(self):
        return "gtk-cancel"

    def get_description(self):
        return _("Mark this task as not to be done anymore")


class CreateNewTask(Action):
    def __init__(self):
        Action.__init__(self, _("Create Task"))

    def activate(self, leaf, iobj=None, ctx=None):
        interface = _create_dbus_connection(True)
        title, body = textutils.extract_title_body(leaf.object)
        interface.OpenNewTask(title, body)

    def item_types(self):
        yield TextLeaf

    def get_icon_name(self):
        return "document-new"

    def get_description(self):
        return _("Create new task in Getting Things GNOME")


class CreateNewEmptyTask(Action):
    def __init__(self):
        Action.__init__(self, _("Create Task"))

    def activate(self, leaf, iobj=None, ctx=None):
        interface = _create_dbus_connection(True)
        interface.OpenNewTask("", "")

    def item_types(self):
        yield AppLeaf

    def valid_for_item(self, leaf):
        return leaf.get_id() == _GTG_ID

    def get_icon_name(self):
        return "document-new"

    def get_description(self):
        return _("Create new task in Getting Things GNOME")


class TasksSource(AppLeafContentMixin, Source):
    appleaf_content_id = _GTG_ID
    source_scan_interval: int = 3600

    def __init__(self, name=None):
        Source.__init__(self, name or __kupfer_name__)
        self._tasks = []
        self._version = 3

    def initialize(self):
        bus = dbus.SessionBus()
        weaklib.dbus_signal_connect_weakly(
            bus,
            "TaskAdded",
            self._on_tasks_updated,
            dbus_interface=_IFACE_NAME2,
        )
        weaklib.dbus_signal_connect_weakly(
            bus,
            "TaskModified",
            self._on_tasks_updated,
            dbus_interface=_IFACE_NAME2,
        )
        weaklib.dbus_signal_connect_weakly(
            bus,
            "TaskDeleted",
            self._on_tasks_updated,
            dbus_interface=_IFACE_NAME2,
        )
        weaklib.dbus_signal_connect_weakly(
            bus,
            "NameOwnerChanged",
            self._name_owner_changed,
            dbus_interface="org.freedesktop.DBus",
            arg0=_SERVICE_NAME2,
        )

    def _name_owner_changed(self, name, old, new):
        if new and not self._tasks:
            self.mark_for_update()

    def _on_tasks_updated(self, task_id):
        self.mark_for_update()

    def get_items(self):
        if (interface := _create_dbus_connection()) is not None:
            self._tasks = list(_load_tasks(interface))
            self._tasks.sort(key=Task.sort_key)

        return self._tasks

    def get_icon_name(self):
        return "gtg"

    def provides(self):
        yield Task



from __future__ import annotations

__kupfer_name__ = _("Screenshot")
__kupfer_sources__ = ("ScreenshotTools",)
__kupfer_text_sources__ = ()
__kupfer_actions__ = ()
__description__ = _("Take screenshot of screen using Scrot or Flameshot.")
__version__ = "2023.1"
__author__ = "KB"

import os
import shutil
import tempfile
import typing as ty
from pathlib import Path

from gi.repository import Gdk, Gtk

from kupfer import icons, launch, plugin_support, runtimehelper
from kupfer.obj import FileLeaf, OperationError, RunnableLeaf, Source
from kupfer.obj.special import CommandNotAvailableLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "tool",
        "label": _("Tool:"),
        "type": str,
        "value": "scrot",
        "alternatives": [
            "Scrot",
            "Flameshot",
        ],
    },
    {
        "key": "format",
        "label": _("Format:"),
        "type": str,
        "value": "jpg",
        "alternatives": [
            "JPG",
            "PNG",
        ],
    },
)


def _tool_cmd_path(tool: str) -> str | None:
    if tool == "Flameshot":
        return shutil.which("flameshot")

    if tool == "Scrot":
        return shutil.which("scrot")

    return None


class ScreenshotToFile(RunnableLeaf):
    def __init__(self, name=_("Take Screenshot To File")):
        super().__init__(name=name)

    def has_result(self):
        return True

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx

        ext = __kupfer_settings__["format"].lower()

        # take temp file name
        file, path = tempfile.mkstemp(f".{ext}", prefix="screenshot")
        os.close(file)
        # file must be deleted
        Path(path).unlink()

        runtimehelper.register_async_file_result(ctx, path)

        tool = __kupfer_settings__["tool"]
        cmd = _tool_cmd_path(tool)
        if not cmd:
            return CommandNotAvailableLeaf(__name__, __kupfer_name__, tool)

        argv: tuple[str, ...]

        if tool == "Flameshot":
            argv = ("flameshot", "gui", "--path", path, "--delay", "1000")
        elif tool == "scrot":
            argv = ("scrot", "--file", path, "--delay", "1")
        else:
            return None

        try:
            launch.spawn_async_raise(argv)
        except launch.SpawnError as exc:
            raise OperationError(exc.args[0].message) from exc

        return FileLeaf(path)

    def item_types(self):
        yield FileLeaf

    def get_gicon(self):
        return icons.ComposedIcon(
            "video-display", "document-save", minimum_icon_size=16
        )


class _SSToClipboardNative(RunnableLeaf):
    """Take screenshot and put image to the clipboard using native method
    provided by tool."""

    def __init__(self, name=_("Take Screenshot to the Clipboard")):
        super().__init__(name=name)

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx

        argv = ("flameshot", "gui", "--clipboard")

        try:
            launch.spawn_async_raise(argv)
        except launch.SpawnError as exc:
            raise OperationError(exc.args[0].message) from exc

    def get_gicon(self):
        return icons.ComposedIcon(
            "video-display", "document-save", minimum_icon_size=16
        )


class SSToClipboard(RunnableLeaf):
    """Take screenshot and put image to the clipboard for tools that not
    provided clipboard support. Take screenshot to file, read image from file and
    put it in the clipboard."""

    def __init__(self, name=_("Take Screenshot to the Clipboard")):
        super().__init__(name=name)

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx

        # take temp file name
        file, path = tempfile.mkstemp(".png", prefix="screenshot")
        os.close(file)

        # for now only scrot
        argv = ["scrot", "--file", path, "--delay", "1", "--overwrite"]

        def finish_callback(acommand, stdout, stderr):
            if not acommand.exit_status and (
                pixbuf := icons.get_pixbuf_from_file(path)
            ):
                clip = Gtk.Clipboard.get_default(Gdk.Display.get_default())
                clip.set_image(pixbuf)

            Path(path).unlink()

        launch.AsyncCommand(argv, finish_callback, 10)

    def get_gicon(self):
        return icons.ComposedIcon(
            "video-display", "document-save", minimum_icon_size=16
        )


class ScreenshotTools(Source):
    serializable = None

    def __init__(self, name=_("Screenshot tools")):
        super().__init__(name)

    def is_dynamic(self):
        return True

    def get_items(self):
        yield ScreenshotToFile()
        tool = __kupfer_settings__["tool"]
        if not _tool_cmd_path(tool):
            yield CommandNotAvailableLeaf(__name__, __kupfer_name__, tool)
            return

        if tool == "Flameshot":
            yield _SSToClipboardNative()
        elif tool == "Scrot":
            yield SSToClipboard()

    def provides(self):
        yield RunnableLeaf

    def get_gicon(self):
        return icons.ComposedIcon(
            "video-display", "document-save", minimum_icon_size=16
        )



from __future__ import annotations

import locale
import typing as ty
from functools import cmp_to_key
from unicodedata import category, normalize

from kupfer.support.datatools import evaluate_once

__all__ = ("get_encoding", "locale_sort", "tofolded", "tolocale", "tounicode")


def _folditems():
    _folding_table = {
        # general non-decomposing characters
        # FIXME: This is not complete
        "ł": "l",
        "œ": "oe",
        "ð": "d",
        "þ": "th",
        # germano-scandinavic canonical transliterations
        "ü": "ue",
        "å": "aa",
        "ä": "ae",
        "æ": "ae",
        "ö": "oe",
        "ø": "oe",
    }

    for char, rep in _folding_table.items():
        yield (ord(char.upper()), rep.upper())
        yield (ord(char), rep)

    yield ord("ß"), "ss"
    yield ord("ẞ"), "SS"


_FOLDING_TABLE = dict(_folditems())


def tounicode(utf8str: ty.Union[str, bytes, None]) -> str | None:
    """Return `unicode` from UTF-8 encoded @utf8str
    This is to use the same error handling etc everywhere
    """
    if isinstance(utf8str, str) or utf8str is None:
        return utf8str

    return utf8str.decode("UTF-8", "replace") if utf8str is not None else ""


# not in use
# def toutf8(ustr: ty.AnyStr) -> bytes:
#     """Return UTF-8 `str` from unicode @ustr
#     This is to use the same error handling etc everywhere
#     if ustr is `str`, just return it
#     """
#     if isinstance(ustr, bytes):
#         return ustr

#     return ustr.encode("UTF-8")


@evaluate_once
def get_encoding() -> str:
    return locale.getpreferredencoding(do_setlocale=False)


def fromlocale(lstr: bytes) -> str:
    """Return a unicode string from locale bytestring @lstr"""
    assert isinstance(lstr, bytes)
    return lstr.decode(get_encoding(), "replace")


def tolocale(ustr: str) -> bytes:
    """Return a locale-encoded bytestring from unicode @ustr"""
    assert isinstance(ustr, str)
    return ustr.encode(get_encoding())


def tofolded(ustr: str) -> str:
    """Fold @ustr

    Return a unicode string where composed characters are replaced by
    their base, and extended latin characters are replaced by
    similar basic latin characters.

    >>> tofolded(u"Wyłącz")
    'Wylacz'
    >>> tofolded(u"naïveté")
    'naivete'
    >>> tofolded(u"Groß")
    'Gross'

    Characters from other scripts are not transliterated.

    >>> print(tofolded(u"Ἑλλάς"))
    Ελλας
    """
    srcstr = normalize("NFKD", ustr.translate(_FOLDING_TABLE))
    return "".join(c for c in srcstr if category(c) != "Mn")


_SortItem = ty.TypeVar("_SortItem")


def locale_sort(
    seq: ty.Iterable[_SortItem], key: ty.Callable[[_SortItem], ty.Any] = str
) -> list[_SortItem]:
    """Return @seq of objects with @key function as a list sorted
    in locale lexical order

    >>> locale.setlocale(locale.LC_ALL, "C")
    'C'
    >>> locale_sort("abcABC")
    ['A', 'B', 'C', 'a', 'b', 'c']

    >>> locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
    'en_US.UTF-8'
    >>> locale_sort("abcABC")
    ['a', 'A', 'b', 'B', 'c', 'C']
    """

    def locale_cmp(val1, val2):
        return locale.strcoll(key(val1), key(val2))

    seq = seq if isinstance(seq, list) else list(seq)
    seq.sort(key=cmp_to_key(locale_cmp))
    return seq


if __name__ == "__main__":
    import doctest

    doctest.testmod()



__kupfer_name__ = _("Prefer Dark Theme")
__description__ = ""
__version__ = "2017.1"
__author__ = ""

import typing as ty

from gi.repository import Gtk

from kupfer import plugin_support
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "prefer_dark",
        "label": _("Prefer Dark Theme"),
        "type": bool,
        "value": True,
    },
)


def initialize_plugin(name):
    use_theme(__kupfer_settings__["prefer_dark"])
    __kupfer_settings__.connect_settings_changed_cb(_on_change_theme)


def finalize_plugin(name):
    use_theme(None)


def _on_change_theme(sender, key, value):
    use_theme(value)


PREFER_DARK = "gtk-application-prefer-dark-theme"


def use_theme(enabled):
    pretty.print_debug(__name__, "updating setting to", enabled)
    sett = Gtk.Settings.get_default()
    if enabled is None:
        sett.reset_property(PREFER_DARK)
    else:
        sett.set_property(PREFER_DARK, enabled)



"""
Changes:
    2018-09-04 * fix: Python 3 compatibility
    2012-10-08 * fix: errors when no one configured directories exists
    2012-06-09 + max depth; fix source name
    2012-06-08 - init
"""

__kupfer_name__ = _("Deep Directories")
__kupfer_sources__ = ("DeepDirSource",)
__description__ = _("Recursive index directories")
__version__ = "2018-09-04"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import os
import typing as ty

from kupfer import plugin_support
from kupfer.obj import apps, files
from kupfer.obj.base import Leaf, Source
from kupfer.obj.filesrc import construct_file_leaf
from kupfer.support import fileutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "dirs",
        "label": _("Directories:"),
        "type": list,
        "helper": "choose_directory",
        "value": ["~/Documents/"],
    },
    {
        "key": "depth",
        "label": _("Depth (max 10):"),
        "type": int,
        "value": 2,
        "max": 10,
        "min": 1,
    },
)


_MAX_DEPTH = 10


class DeepDirSource(Source):
    source_use_cache = False

    def __init__(self, name : str=_("Deep Directories")) -> None:
        super().__init__(name)
        self.dirs : list[str] = []
        self.depth = 1

    def initialize(self) -> None:
        __kupfer_settings__.connect(
            "plugin-setting-changed", self._on_setting_changed
        )
        self.dirs = list(self._get_dirs())
        self.depth = min(__kupfer_settings__["depth"], _MAX_DEPTH)

    def get_items(self) -> ty.Iterable[Leaf]:
        for directory in self.dirs:
            dirfiles = fileutils.get_dirlist(
                directory, max_depth=self.depth, exclude=self._exclude_file
            )
            yield from map(construct_file_leaf, dirfiles)

    def should_sort_lexically(self) -> bool:
        return True

    def _exclude_file(self, filename: str) -> bool:
        return filename.startswith(".")

    def get_description(self) -> str:
        return _("Recursive source of %(dir)s, (%(levels)d levels)") % {
            "dir": ",".join(self.dirs),
            "levels": self.depth,
        }

    def get_icon_name(self) -> str:
        return "folder-saved-search"

    def provides(self) -> ty.Iterator[ty.Type[Leaf]]:
        yield files.FileLeaf
        yield apps.AppLeaf

    @staticmethod
    def _get_dirs() -> ty.Generator[str]:
        for path in __kupfer_settings__["dirs"] or ():
            if (path := os.path.expanduser(path.strip())) and os.path.isdir(
                path
            ):
                yield path

    def _on_setting_changed(self, settings, key, value):
        if key in ("dirs", "depth"):
            self.dirs = list(self._get_dirs())
            self.depth = min(__kupfer_settings__["depth"], _MAX_DEPTH)
            self.mark_for_update()



"""
Implementation of unescaping and unquoting of the Exec= key in
the Desktop Entry Specification (As of March 2011, version 1.1-draft)
http://standards.freedesktop.org/desktop-entry-spec/latest/ar01s06.html
http://standards.freedesktop.org/desktop-entry-spec/desktop-entry-spec-1.1.html#exec-variables

The unescaping we are doing is only one way.. so we unescape according to the
rules, but we accept everything, if validly quoted or not.
"""

from __future__ import annotations

import shlex
import typing as ty

from kupfer.support import itertools as kitertools

__all__ = ("parse_argv", "parse_unesc_argv")

# This is the "string" type encoding escapes
# this is unescaped before we process anything..
_ESCAPE_TABLE = {
    r"\s": " ",
    r"\n": "\n",
    r"\t": "\t",
    r"\r": "\r",
    "\\\\": "\\",
}

# quoted are those chars that need a backslash in front
# (inside a double-quoted section, that is)
# not in use
# quoted = r""" " ` $ \ """.split()
# quoted_table = {
#     r"\"": '"',
#     r"\`": "`",
#     r"\$": "$",
#     "\\\\": "\\",
# }

# reserved are those that need to be inside quotes
# note that all the quoted are also reserved, of course
# We don't use these at all
# reserved = r""" " ' \ > < ~ | & ; $ * ? # ( ) ` """.split()
# reserved.extend([' ', '\t', '\n'])


def _two_part_unescaper(string: str, reptable: dict[str, str]) -> str:
    """Scan @s two characters at a time and replace using @reptable"""
    if not string:
        return string

    def repfunc(instr: str) -> str | None:
        return reptable.get(instr)

    return kitertools.two_part_mapper(string, repfunc)


T = ty.TypeVar("T", str, bytes)


def _custom_shlex_split(
    string: T, comments: bool = False, posix: bool = True
) -> list[T]:
    """Wrapping shlex.split."""
    ustring: str
    if isinstance(string, str):
        ustring = string
    elif isinstance(string, bytes):
        ustring = string.decode("UTF-8", "replace")
    else:
        raise TypeError

    lex = shlex.shlex(ustring, posix=posix)
    lex.whitespace_split = True
    if not comments:
        lex.commenters = ""

    try:
        lex_output = tuple(lex)
    except ValueError:
        lex_output = (ustring,)

    ## extra-unescape  ` and $ that are not handled by shlex
    quoted_shlex = {r"\`": "`", r"\$": "$"}
    output = (_two_part_unescaper(x, quoted_shlex) for x in lex_output)
    if isinstance(string, str):
        return list(output)

    return list(map(str.encode, output))


def _unescape(string: str) -> str:
    """Primary unescape of control sequences"""
    return _two_part_unescaper(string, _ESCAPE_TABLE)


def parse_argv(instr: str) -> list[str]:
    r"""Parse quoted @instr into an argv

    This is according to the spec
    >>> parse_argv('env "VAR=is good" ./program')
    ['env', 'VAR=is good', './program']
    >>> parse_argv('env "VAR=\\\\ \\$ @ x" ./program')
    ['env', 'VAR=\\ $ @ x', './program']
    >>> parse_argv('"\\$" "\\`"  "\\""')
    ['$', '`', '"']
    >>> parse_argv('/usr/bin/x-prog -q %F')
    ['/usr/bin/x-prog', '-q', '%F']
    >>> parse_argv('env LANG=en_US.UTF-8 freeciv-gtk2')
    ['env', 'LANG=en_US.UTF-8', 'freeciv-gtk2']
    >>> parse_argv('emacsclient -a "" -c %f')
    ['emacsclient', '-a', '', '-c', '%f']

    == Below this we need quirks mode ==

    The following style is common but not supported in spec
    >>> parse_argv('env VAR="is broken" ./program')
    ['env', 'VAR=is broken', './program']

    The following is just completely broken
    >>> parse_argv('./program unquoted\\\\argument')
    ['./program', 'unquoted\\argument']

    The following is just completely broken
    >>> parse_argv('./program No\\ Space')
    ['./program', 'No Space']

    The following is just insanely broken
    >>> parse_argv("'/opt'/now/'This is broken/'")
    ['/opt/now/This is broken/']

    This is broken
    #>>> parse_argv('\\$')
    #['$']
    #>>> parse_argv('\\$ \\`  \\"')
    #['$', '`', '"']

    Unmatched quote, normal mode (just testing that it does not raise)
    >>> parse_argv('"hi there')
    ['"hi there']

    Unmatched quote, quirks mode (just testing that it does not raise)
    >>> parse_argv('A\\\\BC "hi there')
    ['A\\\\BC "hi there']

    """
    return _custom_shlex_split(instr)


def parse_unesc_argv(instr: str) -> list[str]:
    r"""Parse quoted @instr into an argv after unescaping it

    >>> parse_unesc_argv(r'stuff "C:\\\\suck\\\\start.exe"')
    ['stuff', 'C:\\suck\\start.exe']

    == Below this we need quirks mode ==

    >>> parse_unesc_argv(r'stuff C:\\\\suck\\\\start.exe')
    ['stuff', 'C:\\suck\\start.exe']

    >>> parse_unesc_argv("'/usr'/bin/gnome-terminal -x gvim 'Insanely Broken'Yes")
    ['/usr/bin/gnome-terminal', '-x', 'gvim', 'Insanely BrokenYes']
    """
    return _custom_shlex_split(_unescape(instr))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import typing as ty

from kupfer.core import commandexec
from kupfer.obj import Leaf, RunnableLeaf, Source

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_sources__ = ("KupferInterals", "CommandResults")
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"


class LastCommand(RunnableLeaf):
    """Represented object is the command tuple to run"""

    qf_id = "lastcommand"

    def __init__(self, obj):
        RunnableLeaf.__init__(self, obj, _("Last Command"))

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx
        obj, action, iobj = self.object
        return ctx.delegated_run(obj, action, iobj)


class KupferInterals(Source):
    def __init__(self):
        Source.__init__(self, _("Internal Kupfer Objects"))

    def is_dynamic(self):
        return True

    def get_items(self):
        ctx = commandexec.default_action_execution_context()
        if ctx.last_command is not None:
            yield LastCommand(ctx.last_command)

    def provides(self):
        yield LastCommand


class LastResultObject(Leaf):
    "dummy superclass"

    def __init__(self, leaf):
        super().__init__(leaf.object, _("Last Result"))


def _make_first_result_object(leaf):
    #    global LastResultObject

    class _LastResultObject(LastResultObject):
        qf_id = "lastresult"

        def __init__(self, leaf):
            super().__init__(leaf)
            vars(self).update(vars(leaf))
            self.name = _("Last Result")
            self.__original_leaf = leaf
            self.__class__.__bases__ = (leaf.__class__, Leaf)

        def get_gicon(self):
            return None

        def get_icon_name(self):
            return Leaf.get_icon_name(self)

        def get_thumbnail(self, width, height):
            return None

        def get_description(self):
            return str(self.__original_leaf)

    return _LastResultObject(leaf)


class CommandResults(Source):
    def __init__(self):
        Source.__init__(self, _("Command Results"))

    def is_dynamic(self):
        return True

    def get_items(self):
        ctx = commandexec.default_action_execution_context()
        yield from reversed(ctx.last_results)
        try:
            leaf = ctx.last_results[-1]
        except IndexError:
            return

        yield _make_first_result_object(leaf)

    def provides(self):
        yield Leaf
        yield LastResultObject



#! /usr/bin/env python3

"""
UI support functions

"""

from gi.repository import Gtk

from kupfer.support import datatools

_escape_table = {
    ord("&"): "&amp;",
    ord("<"): "&lt;",
    ord(">"): "&gt;",
}


def escape_markup_str(mstr: str) -> str:
    """Use a simple homegrown replace table to replace &, <, > with
    entities in @mstr
    """
    return mstr.translate(_escape_table)


@datatools.evaluate_once
def text_direction_is_ltr() -> bool:
    """Check is system is configured as left-to-right."""
    return Gtk.Widget.get_default_direction() != Gtk.TextDirection.RTL  # type: ignore


def normalize_display_name(name: str) -> str:
    # NOT IN USE
    if name[-2] == ":":
        return name + ".0"

    return name


# # NOT IN USE
# def make_rounded_rect(cr, x, y, width, height, radius):
#     """
#     Draws a rounded rectangle with corners of @radius
#     """
#     MPI = math.pi
#     cr.save()

#     cr.move_to(radius, 0)
#     cr.line_to(width-radius,0)
#     cr.arc(width-radius, radius, radius, 3*MPI/2, 2*MPI)
#     cr.line_to(width, height-radius)
#     cr.arc(width-radius, height-radius, radius, 0, MPI/2)
#     cr.line_to(radius, height)
#     cr.arc(radius, height-radius, radius, MPI/2, MPI)
#     cr.line_to(0, radius)
#     cr.arc(radius, radius, radius, MPI, 3*MPI/2)
#     cr.close_path()
#     cr.restore()


# # not in use
# def get_glyph_pixbuf(
#     text: str,
#     size: int,
#     center_vert: bool = True,
#     color: tuple[int, int, int] | None = None,
# ) -> GdkPixbuf.Pixbuf:
#     """Return pixbuf for @text

#     if @center_vert, then center completely vertically
#     """
#     margin = size * 0.1
#     ims = cairo.ImageSurface(cairo.FORMAT_ARGB32, size, size)
#     cctx = cairo.Context(ims)

#     cctx.move_to(margin, size - margin)
#     cctx.set_font_size(size / 2)
#     if color is None:
#         cctx.set_source_rgba(0, 0, 0, 1)
#     else:
#         cctx.set_source_rgb(*color)

#     cctx.text_path(text)
#     x1, y1, x2, y2 = cctx.path_extents()
#     skew_horiz = ((size - x2) - x1) / 2.0
#     skew_vert = ((size - y2) - y1) / 2.0
#     if not center_vert:
#         skew_vert = skew_vert * 0.2 - margin * 0.5

#     cctx.new_path()
#     cctx.move_to(margin + skew_horiz, size - margin + skew_vert)
#     cctx.text_path(text)
#     cctx.fill()

#     ims.flush()
#     pngfile = io.BytesIO()
#     ims.write_to_png(pngfile)

#     loader = GdkPixbuf.PixbufLoader()
#     loader.write(pngfile.getvalue())
#     loader.close()

#     return loader.get_pixbuf()



#! /usr/bin/env python3
# Distributed under terms of the GPLv3 license.

"""
System related functions
"""

import os.path
import socket
import sys

from kupfer.support import datatools, pretty

__all__ = ("get_application_filename", "get_homedir", "get_hostname")


@datatools.evaluate_once
def get_hostname() -> str:
    """Get cached host name"""
    try:
        return socket.gethostname()
    except Exception:
        pretty.print_exc(__name__)

    return ""


@datatools.evaluate_once
def get_homedir() -> str:
    return os.path.expanduser("~")


@datatools.evaluate_once
def get_application_filename() -> str:
    return os.path.basename(sys.argv[0])



"""
A test project to see if we can make a plugin that allows us to
drill down into compressed archives.

So far we only support .zip and .tar, .tar.gz, .tar.bz2, using Python's
standard library.
"""

from __future__ import annotations

__kupfer_name__ = _("Deep Archives")
__kupfer_contents__ = ("ArchiveContent",)
__description__ = _("Allow browsing inside compressed archive files")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import hashlib
import os
import shutil
import tarfile
import typing as ty
import zipfile
from pathlib import Path

from kupfer.obj import FileLeaf, Leaf, Source
from kupfer.obj.filesrc import DirectorySource
from kupfer.support import pretty, scheduler, types

if ty.TYPE_CHECKING:
    from gettext import gettext as _

# Limit this to archives of a couple of megabytes
MAX_ARCHIVE_BYTE_SIZE = 15 * 1024**2

# Wait a year, or until program shutdown for cleaning up
# archive files
VERY_LONG_TIME_S = 3600 * 24 * 365


class UnsafeArchiveError(Exception):
    def __init__(self, path: str):
        Exception.__init__(self, f"Refusing to extract unsafe path: {path}")


def _is_safe_to_unarchive(path: str) -> bool:
    "return whether @path is likely a safe path to unarchive"
    npth = os.path.normpath(path)
    return not os.path.isabs(npth) and not npth.startswith(os.path.pardir)


@ty.runtime_checkable
class _Extractor(ty.Protocol):
    extensions: ty.Collection[str]
    predicate: ty.Callable[[str], bool] | None

    def __call__(self, src: str, dst: str) -> None: ...


class ArchiveContent(Source):
    extractors: ty.ClassVar[list[_Extractor]] = []
    unarchived_files: ty.ClassVar[list[str]] = []
    end_timer = scheduler.Timer(True)

    def __init__(self, fileleaf: FileLeaf, unarchive_func: _Extractor) -> None:
        Source.__init__(self, _("Content of %s") % fileleaf)
        self.path = fileleaf.object
        self.unarchiver = unarchive_func

    def repr_key(self):
        return self.path

    def get_items(self) -> ty.Iterable[Leaf]:
        # always use the same destination for the same file and mtime
        basename = os.path.basename(os.path.normpath(self.path))
        root, _ext = os.path.splitext(basename)
        mtime = os.stat(self.path).st_mtime
        fileid = hashlib.sha1((f"{self.path}{mtime}").encode()).hexdigest()
        pth = os.path.join("/tmp", f"kupfer-{root}-{fileid}")
        if not Path(pth).exists():
            self.output_debug(f"Extracting with {self.unarchiver}")
            self.unarchiver(self.path, pth)
            self.unarchived_files.append(pth)
            self.end_timer.set(
                VERY_LONG_TIME_S, self.clean_up_unarchived_files
            )

        files = list(DirectorySource(pth, show_hidden=True).get_leaves())
        if len(files) == 1 and files[0].has_content():
            csrc = files[0].content_source()
            return (csrc.get_leaves() or []) if csrc else []

        return files

    def get_description(self) -> str | None:
        return None

    @classmethod
    def decorates_type(cls):
        return FileLeaf

    @classmethod
    def decorate_item(cls, leaf: FileLeaf) -> ArchiveContent | None:
        basename = os.path.basename(leaf.object).lower()
        for extractor in cls.extractors:
            if any(basename.endswith(ext) for ext in extractor.extensions):
                assert extractor.predicate
                if Path(leaf.object).is_file() and extractor.predicate(
                    leaf.object
                ):
                    return cls._source_for_path(leaf, extractor)

        return None

    @classmethod
    def _source_for_path(
        cls, leaf: FileLeaf, extractor: _Extractor
    ) -> ArchiveContent | None:
        byte_size = os.stat(leaf.object).st_size
        if byte_size < MAX_ARCHIVE_BYTE_SIZE:
            return cls(leaf, extractor)

        return None

    @classmethod
    def clean_up_unarchived_files(cls) -> None:
        if not cls.unarchived_files:
            return

        def clean_up_error_handler(
            func: ty.Callable[..., object], path: str, exc_info: types.ExecInfo
        ) -> object:
            pretty.print_error(__name__, f"Error in {func} deleting {path}:")
            pretty.print_error(__name__, exc_info)
            return None

        pretty.print_info(__name__, "Removing extracted archives..")
        for filetree in set(cls.unarchived_files):
            pretty.print_debug(
                __name__, "Removing", os.path.basename(filetree)
            )
            # TODO: deprecated: onerror since 3.12
            shutil.rmtree(filetree, onerror=clean_up_error_handler)

        cls.unarchived_files = []

    @classmethod
    def extractor(
        cls,
        extensions: ty.Collection[str],
        predicate: ty.Callable[[str], bool],
    ) -> ty.Callable[[ty.Callable[[str, str], None]], _Extractor]:
        def decorator(func: ty.Callable[[str, str], None]) -> _Extractor:
            extr = ty.cast("_Extractor", func)
            extr.extensions = extensions
            extr.predicate = predicate
            cls.extractors.append(extr)
            return extr

        return decorator


@ArchiveContent.extractor(
    (".tar", ".tar.gz", ".tgz", ".tar.bz2"), tarfile.is_tarfile
)
def extract_tarfile(filepath: str, destpath: str) -> None:
    with tarfile.TarFile.open(filepath, "r") as zfile:
        try:
            for member in zfile.getnames():
                if not _is_safe_to_unarchive(member):
                    raise UnsafeArchiveError(member)

            zfile.extractall(path=destpath)
        finally:
            pass


# ZipFile only supports extractall since Python 2.6
@ArchiveContent.extractor((".zip",), zipfile.is_zipfile)
def extract_zipfile(filepath: str, destpath: str) -> None:
    with zipfile.ZipFile(filepath, "r") as zfile:
        try:
            for member in zfile.namelist():
                if not _is_safe_to_unarchive(member):
                    raise UnsafeArchiveError(member)

            zfile.extractall(path=destpath)
        finally:
            pass



__kupfer_name__ = _("Shell Commands")
__kupfer_sources__ = ()
__kupfer_actions__ = (
    "PassToCommand",
    "FilterThroughCommand",
    "WriteToCommand",
)
__kupfer_text_sources__ = ("CommandTextSource",)
__description__ = (
    _(
        "Run command-line programs. Actions marked with"
        " the symbol %s run in a subshell."
    )
    % "\N{GEAR}"
)
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import typing as ty

from kupfer import icons, launch
from kupfer.obj import (
    Action,
    Execute,
    FileLeaf,
    OperationError,
    TextLeaf,
    TextSource,
)
from kupfer.support import kupferstring, pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _


def finish_command(ctx, acommand, stdout, stderr, post_result=True):
    """Show async error if @acommand returns error output & error status.
    Else post async result if @post_result.
    """
    max_error_msg = 512
    pretty.print_debug(__name__, "Exited:", acommand)
    if acommand.exit_status != 0 and not stdout and stderr:
        errstr = kupferstring.fromlocale(stderr)[:max_error_msg]
        ctx.register_late_error(OperationError(errstr))
    elif post_result:
        leaf = TextLeaf(kupferstring.fromlocale(stdout))
        ctx.register_late_result(leaf)


class GetOutput(Action):
    def __init__(self):
        Action.__init__(self, _("Run (Get Output)"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        if isinstance(leaf, Command):
            argv = ["sh", "-c", leaf.object, "--"]
        else:
            argv = [leaf.object]

        def finish_callback(acommand, stdout, stderr):
            finish_command(ctx, acommand, stdout, stderr)

        pretty.print_debug(__name__, "Spawning with timeout 15 seconds")
        launch.AsyncCommand(argv, finish_callback, 15)

    def get_description(self):
        return _("Run program and return its output") + " \N{GEAR}"


class PassToCommand(Action):
    def __init__(self):
        # TRANS: The user starts a program (command) and the text
        # TRANS: is an argument to the command
        Action.__init__(self, _("Pass to Command..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx
        self.activate_multiple((leaf,), (iobj,), ctx)

    def _run_command(self, objs, iobj, ctx):
        if isinstance(iobj, Command):
            argv = ["sh", "-c", iobj.object + ' "$@"', "--"]
        else:
            argv = [iobj.object]

        def finish_callback(acommand, stdout, stderr):
            finish_command(ctx, acommand, stdout, stderr, False)

        argv.extend([o.object for o in objs])
        pretty.print_debug(__name__, "Spawning without timeout")
        launch.AsyncCommand(argv, finish_callback, None)

    def activate_multiple(self, objs, iobjs, ctx):
        for iobj in iobjs:
            self._run_command(objs, iobj, ctx)

    def item_types(self):
        yield TextLeaf
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf
        yield Command

    def valid_object(self, iobj, for_item=None):
        if isinstance(iobj, Command):
            return True

        return iobj.is_executable()

    def get_description(self):
        return (
            _("Run program with object as an additional parameter")
            + " \N{GEAR}"
        )


class WriteToCommand(Action):
    def __init__(self, name=None):
        # TRANS: The user starts a program (command) and
        # TRANS: the text is written on stdin
        Action.__init__(self, name or _("Write to Command..."))
        self.post_result = False

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx

        if isinstance(iobj, Command):
            argv = ["sh", "-c", iobj.object]
        else:
            argv = [iobj.object]

        def finish_callback(acommand, stdout, stderr):
            finish_command(ctx, acommand, stdout, stderr, self.post_result)

        pretty.print_debug(__name__, "Spawning without timeout")
        output = leaf.object.encode("utf-8")
        launch.AsyncCommand(argv, finish_callback, None, stdin=output)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf
        yield Command

    def valid_object(self, iobj, for_item=None):
        if isinstance(iobj, Command):
            return True

        return iobj.is_executable()

    def get_description(self):
        return (
            _("Run program and supply text on the standard input")
            + " \N{GEAR}"
        )


class FilterThroughCommand(WriteToCommand):
    def __init__(self):
        # TRANS: The user starts a program (command) and
        # TRANS: the text is written on stdin, and we
        # TRANS: present the output (stdout) to the user.
        super().__init__(_("Filter through Command..."))
        self.post_result = True

    def get_description(self):
        return (
            _("Run program and supply text on the standard input")
            + " \N{GEAR}"
        )


class Command(TextLeaf):
    def __init__(self, exepath, name):
        TextLeaf.__init__(self, name, name)
        self.exepath = exepath

    def repr_key(self):
        return self.name

    def get_actions(self):
        yield Execute(quoted=False)
        yield Execute(in_terminal=True, quoted=False)
        yield GetOutput()

    def get_description(self):
        args = " ".join(str(self).split(None, 1)[1:])
        return f"{self.exepath} {args}"

    def get_gicon(self):
        return icons.get_gicon_for_file(self.exepath)

    def get_icon_name(self):
        return "application-x-executable"


class CommandTextSource(TextSource):
    """Yield path and command text items"""

    def __init__(self):
        TextSource.__init__(self, name=_("Shell Commands"))

    def get_rank(self):
        return 80

    def get_text_items(self, text):
        if not text.strip():
            return

        if "\n" in text:
            return

        ## check for absolute path with arguments
        firstword, *rest = text.split(None, 1)
        ## files are handled elsewhere
        if firstword.startswith("/") and not rest:
            return

        ## absolute paths come out here since
        ## os.path.join with two abspaths returns the latter

        # iterate over $PATH directories
        env_path = os.environ.get("PATH", os.defpath)
        for execdir in env_path.split(os.pathsep):
            exepath = os.path.join(execdir, firstword)
            if os.access(exepath, os.R_OK | os.X_OK) and os.path.isfile(
                exepath
            ):
                yield Command(exepath, text)
                break

    def get_description(self):
        return _("Run command-line programs")



__kupfer_name__ = _("Brotab")
__kupfer_sources__ = ("TabSource",)
__kupfer_actions__ = ("TabActivate", "TabClose", "TabGetUrl")
__description__ = _("Firefox Tabs")
__version__ = "2020.1"
__author__ = "Peter Stuifzand <peter@p83.nl>"

import typing as ty

from brotab.api import MultipleMediatorsAPI  # pylint: disable=import-error
from brotab.main import create_clients  # pylint: disable=import-error

from kupfer.obj import Action, Leaf, Source, UrlLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


def get_api():
    clients = create_clients()
    return MultipleMediatorsAPI(clients)


def get_active_tab():
    api = get_api()
    tabs = api.get_active_tabs(None)
    return tabs[0]


def get_tabs():
    api = get_api()
    tabs = api.list_tabs([])
    for tab in tabs:
        yield tab.split("\t")


class TabLeaf(Leaf):
    """
    Tab leaf

    This Leaf represents a tab from Brotab in Firefox
    """

    def __init__(self, tab_id, title, url):
        self.title = title
        self.url = url
        super().__init__(tab_id, title)

    def get_description(self):
        return self.url

    def get_icon_name(self):
        return "web-browser"


class TabSource(Source):
    """
    Tab source

    This Source contains all Tabs from Firefox (as given by Brotab)
    """

    source_use_cache = False
    task_update_interval_sec = 5

    def __init__(self, name=None):
        super().__init__(name or _("Firefox Tabs"))

    def get_items(self):
        cache = {}
        for tab in get_tabs():
            cache[tab[0]] = tab
            yield TabLeaf(tab[0], tab[1], tab[2])

        active_tabs = get_api().get_active_tabs([])
        for active_tab in active_tabs[0]:
            tab = cache[active_tab]
            yield TabLeaf(tab[0], tab[1], tab[2])

    def get_description(self):
        return _("Firefox browser tabs")

    def provides(self):
        yield TabLeaf

    def get_icon_name(self):
        return "web-browser"


class TabActivate(Action):
    def __init__(self):
        super().__init__(_("Activate Tab"))

    def activate(self, leaf, iobj=None, ctx=None):
        get_api().activate_tab([leaf.object], True)

    def item_types(self):
        yield TabLeaf

    def get_icon_name(self):
        return "go-jump"


class TabClose(Action):
    def __init__(self):
        super().__init__(_("Close Tab"))

    def activate(self, leaf, iobj=None, ctx=None):
        get_api().close_tabs([leaf.object])

    def item_types(self):
        yield TabLeaf

    def get_icon_name(self):
        return "window-close"


class TabGetUrl(Action):
    def __init__(self):
        super().__init__(_("Get URL"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        return UrlLeaf(leaf.url, leaf.title)

    def item_types(self):
        yield TabLeaf



"""File-related support functions.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import itertools
import os
import tempfile
import typing as ty
from os import path as os_path
from pathlib import Path

from kupfer.support import pretty

__all__ = (
    "get_destfile",
    "get_destfile_in_directory",
    "get_destpath_in_directory",
    "get_dirlist",
    "get_safe_tempfile",
    "is_directory_writable",
    "is_file_writable",
    "lookup_exec_path",
)


FilterFunc = ty.Callable[[str], bool]


def get_dirlist(
    folder: str,
    max_depth: int = 0,
    include: FilterFunc | None = None,
    exclude: FilterFunc | None = None,
) -> ty.Iterator[str]:
    """Return a list of absolute paths in folder include, exclude: a function
    returning a boolean

    def include(filename):
        return ShouldInclude
    """

    h_include: FilterFunc = include or (lambda _x: True)
    h_exclude: FilterFunc = exclude or (lambda _x: False)

    def accept_file(file):
        return h_include(file) and not h_exclude(file)

    for dirname, dirnames, fnames in os.walk(folder):
        # skip deep directories
        depth = len(os.path.relpath(dirname, folder).split(os.path.sep)) - 1
        if depth >= max_depth:
            # this stop processing subfolders
            dirnames.clear()
            continue

        excl_dir = []
        for directory in dirnames:
            if accept_file(directory):
                yield os_path.join(dirname, directory)
            else:
                excl_dir.append(directory)

        yield from (
            os_path.join(dirname, file) for file in fnames if accept_file(file)
        )

        # do not process excluded dirs
        for directory in reversed(excl_dir):
            dirnames.remove(directory)


def is_directory_writable(dpath: str | Path) -> bool:
    """If directory path @dpath is a valid destination to write new files?"""
    if isinstance(dpath, str):
        dpath = Path(dpath)

    if not dpath.is_dir():
        return False

    return os.access(dpath, os.R_OK | os.W_OK | os.X_OK)


def is_file_writable(dpath: str | Path) -> bool:
    """If @dpath is a valid, writable file

    UNUSED
    """
    if isinstance(dpath, str):
        dpath = Path(dpath)

    if not dpath.is_file():
        return False

    return os.access(dpath, os.R_OK | os.W_OK)


def get_destpath_in_directory(
    directory: str, filename: str, extension: str | None = None
) -> str:
    """Find a good destpath for a file named @filename in path @directory
    Try naming the file as filename first, before trying numbered versions
    if the previous already exist.

    If @extension, it is used as the extension. Else the filename is split and
    the last extension is used
    """
    # find a nonexisting destname
    if extension:
        basename = filename + extension
        root, ext = filename, extension
    else:
        basename = filename
        root, ext = os_path.splitext(filename)

    ctr = itertools.count(1)
    destpath = Path(directory, basename)
    while destpath.exists():
        num = next(ctr)
        basename = f"{root}-{num}{ext}"
        destpath = Path(directory, basename)

    return str(destpath)


def get_destfile_in_directory(
    directory: str, filename: str, extension: str | None = None
) -> tuple[ty.BinaryIO | None, str | None]:
    """Find a good destination for a file named @filename in path @directory.

    Like get_destpath_in_directory, but returns an open file object, opened
    atomically to avoid race conditions.

    Return (fileobj, filepath)
    """
    # retry if it fails
    for _retry in range(3):
        destpath = get_destpath_in_directory(directory, filename, extension)
        try:
            fileno = os.open(
                destpath, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666
            )
        except OSError as exc:
            pretty.print_error(__name__, exc)
        else:
            return (os.fdopen(fileno, "wb"), destpath)

    return (None, None)


def get_destfile(
    destpath: str | Path,
) -> tuple[ty.BinaryIO | None, str | None]:
    """Open file object for full file path. Return the same object
    like get_destfile_in_directory.

    Return (fileobj, filepath).
    """
    # retry if it fails
    for _retry in range(3):
        try:
            fileno = os.open(
                destpath, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666
            )
        except OSError as exc:
            pretty.print_error(__name__, exc)
        else:
            return (os.fdopen(fileno, "wb"), str(destpath))

    return (None, None)


def get_safe_tempfile() -> tuple[ty.BinaryIO, str]:
    """Return (fileobj, filepath) pointing to an open temporary file"""

    fileno, path = tempfile.mkstemp()
    return (os.fdopen(fileno, "wb"), path)


def lookup_exec_path(exename: str) -> str | None:
    """Return path for @exename in $PATH or None"""
    env_path = os.environ.get("PATH") or os.defpath
    for execdir in env_path.split(os.pathsep):
        exepath = Path(execdir, exename)
        if os.access(exepath, os.R_OK | os.X_OK) and exepath.is_file():
            return str(exepath)

    return None



"""
Actions for file leaves.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import os
import typing as ty
from collections import defaultdict

from gi.repository import Gio, GLib

from kupfer import launch
from kupfer.obj import files
from kupfer.obj.base import Action, Leaf
from kupfer.obj.exceptions import NoDefaultApplicationError
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core import commandexec

__all__ = (
    "GetParent",
    "Open",
)


class Open(Action):
    """Open with default application."""

    action_accelerator = "o"
    rank_adjust = 5

    def __init__(self, name: str = _("Open")) -> None:
        Action.__init__(self, name)

    @classmethod
    def default_application_for_leaf(
        cls, leaf: files.FileLeaf
    ) -> Gio.AppInfo | None:
        content_attr = Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE
        gfile = leaf.get_gfile()
        try:
            info = gfile.query_info(
                content_attr, Gio.FileQueryInfoFlags.NONE, None
            )
        except GLib.GError as err:
            pretty.print_error("get appinfo for leaf", leaf, "error", err)
            return None

        content_type = info.get_attribute_string(content_attr)
        def_app = Gio.app_info_get_default_for_type(content_type, False)
        if not def_app:
            raise NoDefaultApplicationError(
                (
                    _("No default application for %(file)s (%(type)s)")
                    % {"file": str(leaf), "type": content_type}
                )
                + "\n"
                + _('Please use "%s"') % _("Set Default Application...")
            )

        return def_app

    def wants_context(self) -> bool:
        return True

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> None:
        assert ctx
        self.activate_multiple((leaf,), ctx)

    def activate_multiple(
        self,
        objects: ty.Iterable[Leaf],
        ctx: commandexec.ExecutionToken | None = None,
    ) -> None:
        appmap: dict[str, Gio.AppInfo] = {}
        leafmap: dict[str, list[files.FileLeaf]] = defaultdict(list)
        for obj in objects:
            assert isinstance(obj, files.FileLeaf)
            if app := self.default_application_for_leaf(obj):
                id_ = app.get_id()
                appmap[id_] = app
                leafmap[id_].append(obj)

        for id_, leaves in leafmap.items():
            app = appmap[id_]
            launch.launch_application(
                app,
                paths=[L.object for L in leaves],
                activate=False,
                screen=ctx and ctx.environment.get_screen(),
            )

    def get_description(self) -> str | None:
        return _("Open with default application")


class GetParent(Action):
    action_accelerator = "p"
    rank_adjust = -5

    def __init__(self, name: str = _("Get Parent Folder")) -> None:
        super().__init__(name)

    def has_result(self) -> bool:
        return True

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> files.FileLeaf:
        assert isinstance(leaf, files.FileLeaf)
        fileloc = leaf.object
        parent = os.path.normpath(os.path.join(fileloc, os.path.pardir))
        return files.FileLeaf(parent)

    def get_description(self) -> str | None:
        return None

    def get_icon_name(self) -> str:
        return "folder-open"



"""
session sets up the program as a client to the current
desktop session and enables notifications on session
close, to be able to save state before being killed;

the module API does not depend on the session API used
"""

from __future__ import annotations

import os
import time
import typing as ty

import dbus
from gi.repository import GObject

from kupfer import version
from kupfer.support import pretty


class SessionClient(GObject.GObject, pretty.OutputMixin):  # type:ignore
    """Session handling controller

    signals:
    save-yourself: Program should save state
    die:           Program should quit immediately
    """

    __gtype_name__ = "SessionClient"

    def __init__(self):
        """Set up program as client to current Session"""
        GObject.GObject.__init__(self)
        self._session_ended = False
        self._client_id = None

        succ = False
        try:
            succ = self._connect_session_manager()
        except dbus.DBusException as exc:
            self.output_error(exc)

        if not succ:
            # try to bind to xfce session manager
            try:
                succ = self._connect_xfce_session_manager()
            except dbus.DBusException as exc:
                self.output_error(exc)

        self._enabled = succ or self._connect_gnomeui()

        # unset autostart id so that it is not transferred
        os.putenv("DESKTOP_AUTOSTART_ID", "")

        if not self.enabled:
            self.output_info(
                "Warning: Not able to connect to current "
                "desktop session, please Quit before logout to save "
                "kupfer's data."
            )

    def _connect_gnomeui(self) -> bool:
        return False

    def _connect_session_manager(self) -> bool:
        bus = dbus.SessionBus()
        proxy_obj = bus.get_object(
            "org.freedesktop.DBus", "/org/freedesktop/DBus"
        )
        dbus_iface = dbus.Interface(proxy_obj, "org.freedesktop.DBus")
        iface_name = service_name = "org.gnome.SessionManager"
        obj_name = "/org/gnome/SessionManager"

        if not dbus_iface.NameHasOwner(service_name):
            self.output_debug("D-Bus name", service_name, "not found")
            return False

        try:
            obj = bus.get_object(service_name, obj_name)
        except dbus.DBusException as exc:
            pretty.print_error(__name__, exc)
            return False

        smanager = dbus.Interface(obj, iface_name)

        app_id = version.PACKAGE_NAME
        startup_id = os.getenv("DESKTOP_AUTOSTART_ID") or ""
        self._client_id = smanager.RegisterClient(app_id, startup_id)
        self._session_ended = False
        self.output_debug(
            "Connected to session as client", self._client_id, startup_id
        )

        private_iface_name = "org.gnome.SessionManager.ClientPrivate"
        bus.add_signal_receiver(
            self._on_query_end_session,
            "QueryEndSession",
            dbus_interface=private_iface_name,
        )
        bus.add_signal_receiver(
            self._on_end_session_signal,
            "EndSession",
            dbus_interface=private_iface_name,
        )
        bus.add_signal_receiver(
            self._on_stop_signal, "Stop", dbus_interface=private_iface_name
        )
        return True

    def _connect_xfce_session_manager(self) -> bool:
        bus = dbus.SessionBus()
        proxy_obj = bus.get_object(
            "org.freedesktop.DBus", "/org/freedesktop/DBus"
        )
        dbus_iface = dbus.Interface(proxy_obj, "org.freedesktop.DBus")
        service_name = "org.xfce.SessionManager"
        obj_name = "/org/xfce/SessionManager"

        if not dbus_iface.NameHasOwner(service_name):
            self.output_debug("D-Bus name", service_name, "not found")
            return False

        try:
            bus.get_object(service_name, obj_name)
        except dbus.DBusException as exc:
            pretty.print_error(__name__, exc)
            return False

        private_iface_name = "org.xfce.Session.Manager"
        bus.add_signal_receiver(
            self._on_xfce_session_state_changed,
            "StateChanged",
            dbus_interface=private_iface_name,
        )
        return True

    def _get_response_obj(self) -> ty.Any:
        """Return D-Bus session object for ClientPrivate Interface"""
        service_name = "org.gnome.SessionManager"
        obj_name = self._client_id
        iface_name = "org.gnome.SessionManager.ClientPrivate"

        try:
            bus = dbus.Bus()
            obj = bus.get_object(service_name, obj_name)
        except dbus.DBusException as exc:
            pretty.print_error(__name__, exc)
            return None

        return dbus.Interface(obj, iface_name)

    def _on_query_end_session(self, flags: str) -> None:
        self.output_debug("Query end", flags)

        if smanager := self._get_response_obj():
            smanager.EndSessionResponse(True, "Always OK")

    def _on_end_session_signal(self, flags: str) -> None:
        self.output_debug("Session end", flags)
        if not self._session_ended:
            self._session_ended = True
            self.emit("save-yourself")

        if smanager := self._get_response_obj():
            smanager.EndSessionResponse(True, "Always OK")

    def _on_xfce_session_state_changed(
        self, old_value: int, new_value: int
    ) -> None:
        self.output_debug(
            "XFCE Session change", time.asctime(), old_value, new_value
        )
        if new_value == 4:  # noqa:PLR2004
            self.emit("save-yourself")

    def _on_stop_signal(self) -> None:
        self.output_debug("Session stop")
        self.emit("die")

    # def _session_save(self, obj: ty.Any, *args: ty.Any) -> None:
    #     self.emit("save-yourself")

    # def _session_die(self, obj: ty.Any, *args: ty.Any) -> None:
    #     self.emit("die")

    @property
    def enabled(self) -> bool:
        """If a session connection is available"""
        return self._enabled


GObject.signal_new(
    "save-yourself",
    SessionClient,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (),
)
GObject.signal_new(
    "die",
    SessionClient,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (),
)



"""
Module provide function to read Thunderbird's address book.

Concept for mork parser from:
    - demork.py by Kumaran Santhanam
    - mork.cs from GnomeDo by Pierre Östlund

2021-01-01 Add support sqlite address book file format
2022-06-14 Support new (?) sqlite address book file format; load also
           history.sqlite
"""

import os
import re
import sqlite3
import time
import typing as ty
from configparser import RawConfigParser
from contextlib import closing

from kupfer.support import pretty

__version__ = "2021-01-01"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

_THUNDERBIRD_HOME = [
    os.path.expanduser(name)
    for name in ("~/.mozilla-thunderbird/", "~/.thunderbird", "~/.icedove/")
]

_THUNDERBIRD_PROFILES = [
    (thome, os.path.join(thome, "profiles.ini")) for thome in _THUNDERBIRD_HOME
]


_RE_COLS = re.compile(r"<\s*<\(a=c\)>\s*(\/\/)?\s*(\(.+?\))\s*>")
_RE_CELL = re.compile(r"\((.+?)\)")
_RE_ATOM = re.compile(r"<\s*(\(.+?\))\s*>")
_RE_TABLE = re.compile(
    r"\{-?(\d+):\^(..)\s*\{\(k\^(..):c\)\(s=9u?\)\s*(.*?)\}\s*(.+?)\}"
)
_RE_ROW = re.compile(r"(-?)\s*\[(.+?)((\(.+?\)\s*)*)\]")
_RE_CELL_TEXT = re.compile(r"\^(.+?)=(.*)")
_RE_CELL_OID = re.compile(r"\^(.+?)\^(.+)")
_RE_TRAN_BEGIN = re.compile(r"@\$\$\{.+?\{\@")
_RE_TRAN_END = re.compile(r"@\$\$\}.+?\}\@")


_COLS_TO_KEEP = (
    "DisplayName",
    "FirstName",
    "LastName",
    "PrimaryEmail",
    "SecondEmail",
)

SPECIAL_CHARS = (
    ("\\\\", "\\"),
    ("\\$", "$"),
    ("\\t", chr(9)),
    ("\\n", chr(10)),
)

RE_ESCAPED = re.compile(r"(\$[a-f0-9]{2})", re.IGNORECASE)
RE_ESCAPEDB = re.compile(rb"(\$[a-f0-9]{2})", re.IGNORECASE)
RE_HEADER = re.compile(r'// <!-- <mdb:mork:z v="(.*)"/> -->')


class _Table:
    def __init__(self, tableid):
        self.tableid = tableid
        self.rows = {}

    def __repr__(self):
        return f"Table {self.tableid!r}: {self.rows!r}"

    def add_cell(self, rowid: str, col: str, atom: str) -> None:
        if ":" in rowid:
            rowid = rowid.split(":")[0]

        row = self.rows.get(rowid)
        if not row:
            row = self.rows[rowid] = {}

        row[col] = _unescape_data(atom)

    def del_row(self, rowid: str) -> None:
        if ":" in rowid:
            rowid = rowid.split(":")[0]

        if rowid in self.rows:
            del self.rows[rowid]


def _unescape_byte(match: re.Match[bytes]) -> bytes:
    value = match.group()
    return bytes([int(value[1:], 16)])


def _unescape_data(instr: str) -> str:
    for src, dst in SPECIAL_CHARS:
        instr = instr.replace(src, dst)

    if RE_ESCAPED.search(instr) is not None:
        inbytes = instr.encode("utf-8")
        instr = RE_ESCAPEDB.sub(_unescape_byte, inbytes).decode(
            "utf-8", "replace"
        )

    return instr


def _read_mork_filecontent(filename: str) -> ty.Iterable[str]:
    with open(filename, encoding="UTF-8") as mfile:
        header = mfile.readline().strip()
        # check header
        if not RE_HEADER.match(header):
            pretty.print_debug(__name__, "_read_mork: header error", header)
            return

        for line in mfile:
            # remove blank lines and comments
            line = line.strip()  # noqa: PLW2901
            if not line:
                continue

            # remove comments
            if (comments := line.find("// ")) > -1:
                line = line[:comments].strip()  # noqa: PLW2901

            if line:
                yield line.replace("\\)", "$29")


# pylint: disable=too-many-locals,too-many-nested-blocks,too-many-branches
# pylint: disable=too-many-statements
def _read_mork(filename: str) -> dict[str, _Table]:  # noqa:PLR0915,PLR0912
    """Read mork file, return tables from file"""

    data = "".join(_read_mork_filecontent(filename))
    if not data:
        return {}

    # decode data
    cells = {}
    atoms = {}
    tables: dict[str, _Table] = {}
    pos = 0
    active_trans = False
    while data:
        data = data[pos:].lstrip()
        if not data:
            break

        # cols
        if match := _RE_COLS.match(data):
            for cell in _RE_CELL.findall(match.group()):
                key, val = cell.split("=", 1)
                if val in _COLS_TO_KEEP:  # skip necessary columns
                    cells[key] = val

            pos = match.span()[1]
            continue

        # atoms
        if match := _RE_ATOM.match(data):
            for cell in _RE_CELL.findall(match.group()):
                if "=" in cell:
                    key, val = cell.split("=", 1)
                    atoms[key] = val

            pos = match.span()[1]
            continue

        # tables
        if match := _RE_TABLE.match(data):
            tableid = ":".join(match.groups()[0:2])
            table = tables.get(tableid)
            if not table:
                table = tables[tableid] = _Table(tableid)

            for row in _RE_ROW.findall(match.group()):
                tran, rowid = row[:2]
                if active_trans and rowid[0] == "-":
                    rowid = rowid[1:]
                    table.del_row(rowid)

                if not active_trans or tran != "-":
                    rowdata = row[2:]
                    rowcell: str
                    for rowcell in filter(None, rowdata):
                        for cell in _RE_CELL.findall(rowcell):
                            atom, col = None, None
                            if cmatch := _RE_CELL_TEXT.match(cell):
                                col = cells.get(cmatch.group(1))
                                atom = cmatch.group(2)
                            elif cmatch := _RE_CELL_OID.match(cell):
                                col = cells.get(cmatch.group(1))
                                atom = atoms.get(cmatch.group(2))

                            if col and atom:
                                table.add_cell(rowid, col, atom)

            pos = match.span()[1]
            continue

        # transaction
        if _RE_TRAN_BEGIN.match(data):
            active_trans = True
            continue

        if _RE_TRAN_END.match(data):
            tran = True
            continue

        # dangling rows
        if match := _RE_ROW.match(data):
            row = match.groups()
            tran, rowid = row[:2]
            table = tables.get("1:80")  # bind to default table
            if rowid[0] == "-":
                rowid = rowid[1:]
                if table:
                    table.del_row(rowid)

            if tran != "-" and (rowdata := row[2:]):
                if not table:
                    table = tables["1:80"] = _Table("1:80")

                for rowcell in filter(None, rowdata):
                    for cell in _RE_CELL.findall(rowcell):
                        atom, col = None, None
                        if cmatch := _RE_CELL_TEXT.match(cell):
                            col = cells.get(cmatch.group(1))
                            atom = cmatch.group(2)
                        elif cmatch := _RE_CELL_OID.match(cell):
                            col = cells.get(cmatch.group(1))
                            atom = atoms.get(cmatch.group(2))

                        if col and atom:
                            table.add_cell(rowid, col, atom)

            pos = match.span()[1]
            continue

        pos = 1

    return tables


def _mork2contacts(tables: dict[str, _Table]) -> ty.Iterator[tuple[str, str]]:
    """Get contacts from mork table prepared by _read_mork"""
    if not tables:
        return
    # get only default table
    if table := tables.get("1:80"):
        for row in table.rows.values():
            display_name = row.get("DisplayName")
            if not display_name:
                first_name = row.get("FirstName", "")
                last_name = row.get("LastName", "")
                display_name = " ".join((first_name, last_name))

            if display_name:
                display_name = display_name.strip()

            for key in ("PrimaryEmail", "SecondEmail"):
                if email := row.get(key):
                    yield (display_name or email[: email.find("@")], email)


_ABOOK_CONTACTS_SQL = """
select
    (select value from properties
     where card = c.uid and name = 'FirstName'
    ) as FirstName,
    (select value from properties
     where card = c.uid and name = 'LastName'
    ) as LastName,
    (select value from properties
     where card = c.uid and name = 'DisplayName'
    ) as DisplayName,
    (select value from properties
     where card = c.uid and name = 'PrimaryEmail'
    ) as PrimaryEmail,
    (select value from properties
     where card = c.uid and name = 'SecondEmail'
    ) as SecondEmail
from cards c
"""

# new version of abook.sqlite file
_ABOOK_CONTACTS_SQL2 = """
select
    (select value from properties
     where card = c.card and name = 'FirstName'
    ) as FirstName,
    (select value from properties
     where card = c.card and name = 'LastName'
    ) as LastName,
    (select value from properties
     where card = c.card and name = 'DisplayName'
    ) as DisplayName,
    (select value from properties
     where card = c.card and name = 'PrimaryEmail'
    ) as PrimaryEmail,
    (select value from properties
     where card = c.card and name = 'SecondEmail'
    ) as SecondEmail
from (select distinct card from properties) c
"""


def _load_abook_sqlite(filename: str) -> ty.Iterator[tuple[str, str]]:
    """Load contacts from abook.sqlite filename.

    Thunderbird (like firefox) lock database, so it mus be opened as immutable
    and read-only. Also changes may be not visible immediate - require close
    sqlite file by thunderbird.
    """

    dbfpath = filename.replace("?", "%3f").replace("#", "%23")
    dbfpath = "file:" + dbfpath + "?immutable=1&mode=ro"

    for _ in range(2):
        try:
            pretty.print_debug(__name__, "_load_abook_sqlite load:", dbfpath)
            with closing(
                sqlite3.connect(dbfpath, uri=True, timeout=1)
            ) as conn:
                cur = conn.cursor()

                # check db version
                cur.execute(
                    "select count(*) from sqlite_schema "
                    "where name = 'list_cards'"
                )
                ver = cur.fetchone()[0]
                abook_query = (
                    _ABOOK_CONTACTS_SQL2 if ver else _ABOOK_CONTACTS_SQL
                )
                cur.execute(abook_query)

                for (
                    first_name,
                    last_name,
                    display_name,
                    primary_email,
                    second_email,
                ) in cur:
                    display_name = display_name or " ".join(  # noqa: PLW2901
                        filter(None, (first_name, last_name))
                    )
                    for email in (primary_email, second_email):
                        if email:
                            yield (
                                display_name or email.partition("@")[0],
                                email,
                            )
            return
        except sqlite3.Error as err:
            # Something is wrong with the database
            # wait short time and try again
            pretty.print_debug(__name__, "_load_abook_sqlite error:", str(err))
            time.sleep(1)


def get_addressbook_dirs() -> ty.Iterator[str]:
    """Get path to addressbook file from default profile."""
    for thome, tprofile in _THUNDERBIRD_PROFILES:
        if os.path.isfile(tprofile):
            config = RawConfigParser()
            config.read(tprofile)
            for section in config.sections():
                if config.has_option(section, "Path"):
                    path = config.get(section, "Path")
                    if not os.path.isabs(path):
                        path = os.path.join(thome, path)

                    if os.path.isdir(path):
                        yield path


def get_addressbook_files() -> ty.Iterator[str]:
    """Get full path to all Thunderbird address book files."""
    for path in get_addressbook_dirs():
        pretty.print_debug(__name__, "get_addressbook_files dir:", path)
        with os.scandir(path) as entries:
            for entry in entries:
                if not entry.is_file():
                    continue

                if (
                    entry.name.endswith(".mab")
                    or (
                        entry.name.endswith(".sqlite")
                        and entry.name.startswith("abook")
                    )
                    or entry.name == "history.sqlite"
                ):
                    yield entry.path


def get_contacts() -> ty.Iterator[tuple[str, str]]:
    """Get all contacts from all Thunderbird address books as
    ((contact name, contact email))"""
    for abook in get_addressbook_files():
        pretty.print_debug(__name__, "get_contacts:", abook)
        if abook.endswith(".sqlite"):
            try:
                yield from _load_abook_sqlite(abook)
            except Exception as err:
                pretty.print_error(__name__, "get_contacts error", abook, err)
        else:
            try:
                tables = _read_mork(abook)
            except OSError as err:
                pretty.print_error(__name__, "get_contacts error", abook, err)
            else:
                yield from _mork2contacts(tables)


if __name__ == "__main__":
    print("\n".join(map(str, sorted(get_contacts()))))



from __future__ import annotations

import sys
import threading
import time
import typing as ty

from gi.repository import GLib

from kupfer.support import pretty, scheduler

if ty.TYPE_CHECKING:
    from kupfer.support.types import ExecInfo

__all__ = ("Task", "TaskRunner", "ThreadTask")

TaskCallback = ty.Callable[[ty.Any], None]


class Task:
    """Represent a task that can be done in the background

    The finish_callback received in Task.start(..) must be stored,
    and regardless if the task exits with an error, or completes
    successfully, the callback *must* be called.

    The finish callback must pass the Task instance itself as
    the only and first argument:

        finish_callback(self)
    """

    def __init__(self, name: str | None = None) -> None:
        self.name = name

    def __repr__(self) -> str:
        name = repr(getattr(self, "name", None))
        return f"<{type(self).__module__}.{type(self).__name__} name={name}>"

    def start(self, finish_callback: TaskCallback) -> None:
        raise NotImplementedError


class ThreadTask(Task):
    """Run in a thread.

    `wait_sec` (when >0) specify time when we wait for result. After this
    time `thread_finish`, `finish_callback` and `thread_finally` are not called.
    We can't simply kill background thread but we can ignore its result and
    log errors (if any).
    """

    def __init__(self, name: str | None = None, wait_sec: int = 0):
        Task.__init__(self, name)
        self._finish_callback: TaskCallback | None = None
        # what long we wait for result; after this time _finish_callback is not
        # called
        self.wait_sec = wait_sec
        self.thread_started_ts: float = 0

    def thread_do(self) -> None:
        """Override this to run what should be done in the thread"""
        raise NotImplementedError

    def thread_finish(self) -> None:
        """This finish function runs in the main thread after thread
        completion, and can be used to communicate with the GUI.
        """

    def thread_finally(self, exc_info: ExecInfo | None) -> None:
        """Always run at thread finish"""
        if exc_info is not None:
            _etype, value, tback = exc_info
            if value:
                raise value.with_traceback(tback)

    def _thread_finally(self, exc_info: ExecInfo | None) -> None:
        try:
            self.thread_finally(exc_info)
        finally:
            if self._finish_callback:
                self._finish_callback(self)

    def _log_error(self, exc_info: ExecInfo | None) -> None:
        if exc_info:
            pretty.print_exc(__name__, exc_info)

    def _run_thread(self) -> None:
        exc_info = None
        overdue = False
        try:
            self.thread_do()
            overdue = self.wait_sec > 0 and (
                time.monotonic() - self.thread_started_ts > self.wait_sec
            )
            if not overdue:
                GLib.idle_add(self.thread_finish)

        except Exception:
            exc_info = sys.exc_info()
        finally:
            if overdue:
                GLib.idle_add(self._log_error, exc_info)
            else:
                GLib.idle_add(self._thread_finally, exc_info)

    def start(self, finish_callback: TaskCallback) -> None:
        self._finish_callback = finish_callback
        self.thread_started_ts = time.monotonic()
        thread = threading.Thread(target=self._run_thread)
        thread.start()


class TaskRunner(pretty.OutputMixin):
    """Run Tasks in the idle Loop"""

    def __init__(self, end_on_finish: bool) -> None:
        self.tasks: set[Task] = set()
        self.end_on_finish = end_on_finish
        scheduler.get_scheduler().connect("finish", self._on_finish)

    def _task_finished(self, task):
        self.output_debug("Task finished", task)
        self.tasks.remove(task)

    def add_task(self, task: Task) -> None:
        """Register @task to be run"""
        self.tasks.add(task)
        task.start(self._task_finished)

    def _on_finish(self, _sched: ty.Any) -> None:
        if self.end_on_finish:
            self.tasks.clear()
            return

        if self.tasks:
            self.output_info("Uncompleted tasks:")
            for task in self.tasks:
                self.output_info(task)



# Distributed under terms of the GPLv3 license.

"""
Custom widgets used mostly in Preferences Dialog
"""

from __future__ import annotations

import os
import re
import traceback
import typing as ty

from gi.repository import Gio, GLib, Gtk

from kupfer import icons, launch, plugin_support
from kupfer.core import plugins, settings
from kupfer.obj import KupferObject, Source

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.support import types as kty


# pylint: disable=too-few-public-methods
class DirsSelectWidget(Gtk.Bin):  # type: ignore
    """Widgets with list of folders and buttons to add/remove folder into."""

    def __init__(
        self,
        plugin_id: str,
        setting: str,
        plugin_settings: plugin_support.PluginSettings,
    ):
        """Create widgets.
        Args:
            `plugin_id`: plugin id
            `setting`: name of parameter
            `plugin_support`: object PluginSettings for given plugin
        """
        super().__init__()

        self.plugin_id = plugin_id
        self.setting = setting
        self.plugin_settings = plugin_settings

        self.model = Gtk.ListStore.new([str, Gio.Icon, str])
        self.view = Gtk.TreeView.new_with_model(self.model)

        # pylint: disable=no-member
        self.btn_add = Gtk.Button.new_from_icon_name(
            "add", Gtk.IconSize.BUTTON
        )
        self.btn_add.always_show_image = True
        self.btn_add.set_label(_("Add"))
        # pylint: disable=no-member
        self.btn_del = Gtk.Button.new_from_icon_name(
            "remove", Gtk.IconSize.BUTTON
        )
        self.btn_del.always_show_image = True
        self.btn_del.set_label(_("Remove"))

        self._configure_model()
        self._create_layout()

        if dirs := plugin_settings[setting]:
            self._add_dirs(dirs)

        self.view.connect("cursor-changed", self._on_cursor_changed)
        self.btn_del.connect("clicked", self._on_del_clicked)
        self.btn_add.connect("clicked", self._on_add_clicked)

    def _configure_model(self):
        view = self.view

        view.set_headers_visible(False)
        view.set_property("enable-search", False)
        view.get_selection().set_mode(Gtk.SelectionMode.BROWSE)

        icon_cell = Gtk.CellRendererPixbuf()
        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "gicon", 1)

        cell = Gtk.CellRendererText()
        setctl = settings.get_settings_controller()
        cell.set_property("ellipsize", setctl.get_ellipsize_mode())

        col = Gtk.TreeViewColumn("name", cell)
        col.add_attribute(cell, "text", 2)

        view.append_column(icon_col)
        view.append_column(col)
        view.show()

    def _create_layout(self):
        box = Gtk.VBox()
        box.set_spacing(3)

        scrollwin = Gtk.ScrolledWindow()
        # pylint: disable=no-member
        scrollwin.set_shadow_type(type=Gtk.ShadowType.IN)
        scrollwin.set_hexpand(True)
        scrollwin.set_size_request(50, 100)
        # pylint: disable=no-member
        scrollwin.add(self.view)

        box.pack_start(scrollwin, True, True, 0)

        # buttons
        bbox = Gtk.HButtonBox()
        bbox.set_property("layout-style", Gtk.ButtonBoxStyle.END)
        bbox.pack_start(self.btn_del, False, False, 0)
        bbox.pack_start(self.btn_add, False, False, 0)
        box.pack_end(bbox, True, True, 0)

        box.set_hexpand(True)
        self.add(box)

    def _add_dirs(self, dirs: list[str]) -> None:
        for directory in map(os.path.expanduser, dirs):
            dispname = launch.get_display_path_for_bytestring(directory)
            gicon = icons.get_gicon_for_file(directory)
            self.model.append((directory, gicon, dispname))

    def _on_cursor_changed(self, table: Gtk.TreeView) -> None:
        curpath, _curcol = table.get_cursor()
        self.btn_del.set_sensitive(bool(curpath))

    def _on_del_clicked(self, widget: Gtk.Widget) -> None:
        curpath, _curcol = self.view.get_cursor()
        if curpath:
            rowiter = self.model.get_iter(curpath)
            self.model.remove(rowiter)
            self._save()

    def _on_add_clicked(self, widget: Gtk.Widget) -> None:
        # TRANS: File Chooser Title
        chooser_dialog = Gtk.FileChooserDialog(
            title=_("Choose a Directory"),
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.REJECT,
                Gtk.STOCK_OK,
                Gtk.ResponseType.ACCEPT,
            ),
        )
        chooser_dialog.set_select_multiple(True)

        # pylint: disable=no-member
        if chooser_dialog.run() == Gtk.ResponseType.ACCEPT:
            # pylint: disable=no-member
            have = self.current_dirs()
            new_dirs = [
                directory
                for directory in chooser_dialog.get_filenames()
                if directory not in have
            ]
            self._add_dirs(new_dirs)
            self._save()

        chooser_dialog.hide()

    def current_dirs(self) -> list[str]:
        return [os.path.normpath(row[0]) for row in self.model]

    def _save(self) -> None:
        setctl = settings.get_settings_controller()
        setctl.set_plugin_config(
            self.plugin_id, self.setting, self.current_dirs(), list
        )


# pylint: disable=too-few-public-methods
class FileDirSelectWidget(Gtk.Bin):  # type: ignore
    """Widgets with entry and buttons that allow to select file or directory."""

    def __init__(
        self,
        plugin_id: str,
        setting: str,
        plugin_settings: plugin_support.PluginSettings,
        kind: str,
    ):
        """Create widgets.
        Args:
            `plugin_id`: plugin id
            `setting`: name of parameter
            `plugin_support`: object PluginSettings for given plugin
            `kind`: what user can select via dialog: 'choose_file' or
                'choose_directory'
        """
        assert kind in ("choose_directory", "choose_file")

        super().__init__()
        self.plugin_id = plugin_id
        self.setting = setting
        self.plugin_settings = plugin_settings
        self.kind = kind

        self.entry = Gtk.Entry()

        self.btn_add = Gtk.Button.new_from_icon_name(
            "fileopen",
            Gtk.IconSize.BUTTON,  # pylint: disable=no-member
        )
        self.btn_add.always_show_image = True
        self.btn_add.set_label(_("Select"))
        self.btn_clear = Gtk.Button.new_from_icon_name(
            "editclear",
            Gtk.IconSize.BUTTON,  # pylint: disable=no-member
        )
        self.btn_clear.always_show_image = True
        self.btn_clear.set_label(_("Clear"))

        self._create_layout()

        self.entry.set_text(plugin_settings[setting])

        self.btn_clear.connect("clicked", self._on_del_clicked)
        self.btn_add.connect("clicked", self._on_add_clicked)
        self.entry.connect("changed", self._on_change)

    def _create_layout(self):
        box = Gtk.VBox()
        box.set_spacing(3)

        box.pack_start(self.entry, True, True, 0)

        # buttons
        bbox = Gtk.HButtonBox()
        bbox.set_property("layout-style", Gtk.ButtonBoxStyle.END)
        bbox.pack_start(self.btn_clear, False, False, 0)
        bbox.pack_start(self.btn_add, False, False, 0)
        box.pack_end(bbox, True, True, 0)

        box.set_hexpand(True)
        self.add(box)

    def _on_change(self, widget: Gtk.Widget) -> None:
        setctl = settings.get_settings_controller()
        setctl.set_plugin_config(
            self.plugin_id, self.setting, widget.get_text(), str
        )

    def _on_del_clicked(self, widget: Gtk.Widget) -> None:
        self.entry.set_text("")

    def _on_add_clicked(self, widget: Gtk.Widget) -> None:
        # TRANS: File Chooser Title
        if self.kind == "choose_directory":
            title = _("Choose a Directory")
            action = Gtk.FileChooserAction.SELECT_FOLDER
        else:
            title = _("Choose a File")
            action = Gtk.FileChooserAction.OPEN

        chooser_dialog = Gtk.FileChooserDialog(
            title=title,
            action=action,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.REJECT,
                Gtk.STOCK_OK,
                Gtk.ResponseType.ACCEPT,
            ),
        )
        chooser_dialog.set_select_multiple(False)
        if fname := self.entry.get_text():
            fname = os.path.expanduser(fname)
            chooser_dialog.set_filename(fname)  # pylint: disable=no-member

        # pylint: disable=no-member
        if chooser_dialog.run() == Gtk.ResponseType.ACCEPT:
            # pylint: disable=no-member
            fname = chooser_dialog.get_filename()
            fname = os.path.normpath(fname)
            self.entry.set_text(fname)

        chooser_dialog.hide()


# pylint: disable=too-few-public-methods
class ObjectsInfoWidget(Gtk.Bin):  # type: ignore
    """Widget with list of `objs` (sources, actions): name, description."""

    def __init__(
        self,
        plugin_id: str,
        objs: ty.Iterable[str],
    ) -> None:
        super().__init__()

        setctl = settings.get_settings_controller()
        small_icon_size = setctl.get_config_int(
            "Appearance", "icon_small_size"
        )

        box = Gtk.Grid()
        box.set_row_spacing(6)
        box.set_column_spacing(12)

        for row, item in enumerate(objs):
            plugin_type = plugins.get_plugin_attribute(plugin_id, item)
            if not plugin_type:
                continue

            obj: KupferObject = plugin_type()

            # object icon
            image = Gtk.Image()
            image.set_property("gicon", obj.get_icon())
            image.set_property("pixel-size", small_icon_size)
            image.set_alignment(0, 0)  # pylint: disable=no-member
            box.attach(image, 0, row, 1, 1)

            ibox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 3)
            # name and description
            ibox.pack_start(self._create_label(obj), False, True, 0)
            box.attach(ibox, 1, row, 1, 1)

            # Display information for application content-sources.
            # only sources have leaf representation
            if isinstance(obj, Source):
                _valid, leaf_repr = obj.get_valid_leaf_repr()
                if leaf_repr is not None:
                    hbox = self._create_leaves_info(leaf_repr, small_icon_size)
                    ibox.pack_start(hbox, True, True, 0)

        self.add(box)

    def _create_label(self, obj: KupferObject) -> Gtk.Label:
        name_label = GLib.markup_escape_text(str(obj))  # name
        if desc := GLib.markup_escape_text(obj.get_description() or ""):
            name_label = f"{name_label}\n<small>{desc}</small>"

        label = Gtk.Label()
        label.set_alignment(0, 0)  # pylint: disable=no-member
        label.set_markup(name_label)
        label.set_line_wrap(True)  # pylint: disable=no-member
        label.set_selectable(True)
        return label

    def _create_leaves_info(
        self, leaf_repr: KupferObject, small_icon_size: int
    ) -> Gtk.Box:
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 6)
        image = Gtk.Image()
        image.set_property("gicon", leaf_repr.get_icon())
        image.set_property("pixel-size", small_icon_size // 2)
        hbox.pack_start(Gtk.Label.new(_("Content of")), False, True, 0)
        hbox.pack_start(image, False, True, 0)
        hbox.pack_start(Gtk.Label.new(str(leaf_repr)), False, True, 0)
        return hbox


# pylint: disable=too-few-public-methods
class PluginAboutWidget(Gtk.Bin):  # type: ignore
    """Widget with basic information about plugin."""

    def __init__(self, plugin_id: str, info: dict[str, ty.Any]) -> None:
        super().__init__()

        box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 12)

        ver, description, author = plugins.get_plugin_attributes(
            plugin_id,
            (
                plugins.PluginAttr.VERSION,
                plugins.PluginAttr.DESCRIPTION,
                plugins.PluginAttr.AUTHOR,
            ),
        )
        infobox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 6)

        # TRANS: Plugin info fields
        for field, val in (
            (_("Description"), description),
            (_("Author"), author),
            (_("Version"), ver),
        ):
            if sbox := self._create_title_val_label_box(field, val):
                infobox.pack_start(sbox, False, True, 0)

        box.pack_start(infobox, False, True, 0)

        errormsg = None
        # Check for plugin load exception
        if (exc_info := plugins.get_plugin_error(plugin_id)) is not None:
            errstr = _format_exc_info(exc_info)
            errormsg = "".join(
                (
                    "<b>",
                    _("Plugin could not be read due to an error:"),
                    "</b>\n\n",
                    GLib.markup_escape_text(errstr),
                )
            )
        elif not plugins.is_plugin_loaded(plugin_id):
            errormsg = (
                "<i>"
                + GLib.markup_escape_text(_("Plugin is disabled"))
                + "</i>"
            )

        if errormsg:
            label = Gtk.Label()
            label.set_markup(errormsg)
            label.set_alignment(0, 0)  # pylint: disable=no-member
            label.set_line_wrap(True)  # pylint: disable=no-member
            label.set_selectable(True)
            box.pack_start(label, False, True, 0)

        self.add(box)

    def _create_title_val_label_box(
        self, title: str, value: str
    ) -> Gtk.Box | None:
        if not value:
            return None

        sbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 3)
        for label_text in (f"<b>{title}</b>", GLib.markup_escape_text(value)):
            label = Gtk.Label()
            label.set_markup(label_text)
            label.set_alignment(0, 0)  # pylint: disable=no-member
            label.set_line_wrap(True)  # pylint: disable=no-member
            label.set_selectable(True)
            sbox.pack_start(label, False, True, 0)

        return sbox


def _format_exc_info(exc_info: kty.ExecInfo) -> str:
    """Format ExecInfo to presentable form."""
    etype, error, _tb = exc_info
    import_error_pat = r"No module named ([^\s]+)"
    errmsg = str(error)

    if re.match(import_error_pat, errmsg):
        # TRANS: Error message when Plugin needs a Python module to load
        import_error_localized = _("Python module '%s' is needed") % "\\1"
        return re.sub(
            import_error_pat, import_error_localized, errmsg, count=1
        )

    if etype and issubclass(etype, ImportError):
        return errmsg

    return "".join(traceback.format_exception(*exc_info))


def new_label_header(
    parent: Gtk.Box | None, markup: str, tooltip: str | None = None
) -> Gtk.Label:
    """Return "header" label with `markup` text and optional `tooltip`.
    Add it to `parent` box if given. Return new label object."""
    markup = GLib.markup_escape_text(markup)
    return new_label(
        parent, f"<b>{markup}</b>", selectable=False, tooltip=tooltip
    )


def new_label(  # pylint: disable=keyword-arg-before-vararg
    parent: Gtk.Box | None = None,
    /,
    *markup: str,
    selectable: bool = True,
    tooltip: str | None = None,
) -> Gtk.Label:
    """Create new label with `markup` text (concatenated).
    Set optional parameters: `selectable` (default True), `tooltip`.
    If `parent` box is given, add label to box.
    Return widget object."""
    text = "".join(markup)
    label = Gtk.Label()
    label.set_alignment(0, 0.5)  # pylint: disable=no-member
    label.set_markup(text)
    label.set_line_wrap(True)  # pylint: disable=no-member
    label.set_selectable(selectable)

    if tooltip:
        label.set_tooltip_text(tooltip)

    if parent:
        parent.pack_start(label, False, True, 0)

    return label
